#
# Makefile.PL - Automatic Makefile generator for ZETA Perl
#
# ZETA Perl - Perl binding for the ANSI/NISO Z39.50-1995 Protocol
# Copyright (c) 1996, 1999 Finsiel S.p.A.
#
# Rocco Carbone <rocco@tecsiel.it> 4Q 1996
# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

BEGIN { require 5.003 }

use ExtUtils::MakeMaker;
use Config;

# See lib/ExtUtils/MakeMaker.pm for details of how to influence
# the contents of the Makefile that is written.

my $OpSys = $^O;
my $OpSysVer = $Config{osvers};

if ($] < 5.004)
{
  print "\n*** You are using an old version of Perl. ***\n\a";
  print "Perl 5.004 fixes many bugs in older versions and is highly recommended.\n";
  print "You may see warnings from 'pod2man' due to the old version you have.\n";
  print "You can ignore them, or, better still, upgrade to the latest release\n";
  print "of Perl (5.004_04 at the time this was written).\n\n";
  sleep 5;
}


warn <<EOW;

  Remember to read the README file!

  Warning:
     The ZETA perl module is ALPHA software.  It is __only__ 'alpha'
     because its interface is not standardized.  The 'alpha' status
     does not refer to code quality or stability.

  Use `make test' to execute self tests.
  Use `make testdb' to start the tests defined in test.pl under perl debugger.

  Use `make install' to install the ZETA Perl package.
  This will install the necessary files in your Perl directory.
  You may need to have root previliges in order to do so.

  Execute 'perldoc ZETA' after installing to have access to the documentation.

EOW


warn "WARNING: Your GNU C compiler is very old. Please upgrade.\n"
    if ($Config{gccversion} and $Config{gccversion} =~ m/^(1|2\.[1-5])/);


# /NODEFAULTLIB libcmt.lib'
#     'distclean'    => { FILES => "Makefile.old" },

&WriteMakefile(
    'NAME'         => 'ZETA',
    'VERSION_FROM' => 'ZETA.pm',
    'OBJECT'       => 'ZETA$(OBJ_EXT) zutils$(OBJ_EXT) argv$(OBJ_EXT) lib/libzeta$(LIB_EXT) lib/libtelsil$(LIB_EXT)',
    'LIBS'         => ['-lm'],
    'DEFINE'       => '',
    'INC'          => '-Ilib',
    'PM'           => {'ZETA.pm' => '$(INST_LIBDIR)/ZETA.pm'},
    'clean'        => { FILES => "Makefile.old" },
    XSPROTOARG     => '-noprototypes',
    MAN3PODS       => ' ',               # Pods will be built by installman.
    'dist'         => { COMPRESS => 'gzip', SUFFIX => 'gz'},
);


sub MY::post_initialize
{
  my ($self) = shift;

  if (-e "zetaperlrc")
    {
      print "
  In order to test the functionality of this module, you need to have
  a reachable Z39.50 target.  For this reason, a System Wide configuration
  file with default values is distributed with the package.  You may choose
  your prefered values later.  However, if you are not sure what you are doing,
  or if you do not have access to a Z39.50 target, please leave these
  parameters to their respective default values.

   ";

      '';
    }
  else
    {
      print "
  <<<<<< WARNING
  Could not locate the system configuration file. Your package seems to be
  incomplete. However, this may not be harmful.  Try <make test> to regenerate
  the system configuration file.  If the problem continue to exist, please
  inform the author.
  >>

   ";
      '';
    }
}


sub MY::postamble
  {
"
   test1 : pure_all
\tPERL_DL_NONLAZY=1 \$(FULLPERL) -I\$(INST_ARCHLIB) -I\$(INST_LIB) -I\$(PERL_ARCHLIB) -I\$(PERL_LIB) t/1.connect.t

   test2 : pure_all
\tPERL_DL_NONLAZY=1 \$(FULLPERL) -I\$(INST_ARCHLIB) -I\$(INST_LIB) -I\$(PERL_ARCHLIB) -I\$(PERL_LIB) t/2.init.t

   test3 : pure_all
\tPERL_DL_NONLAZY=1 \$(FULLPERL) -I\$(INST_ARCHLIB) -I\$(INST_LIB) -I\$(PERL_ARCHLIB) -I\$(PERL_LIB) t/3.search.t

   test4 : pure_all
\tPERL_DL_NONLAZY=1 \$(FULLPERL) -I\$(INST_ARCHLIB) -I\$(INST_LIB) -I\$(PERL_ARCHLIB) -I\$(PERL_LIB) t/4.present.t

   test5 : pure_all
\tPERL_DL_NONLAZY=1 \$(FULLPERL) -I\$(INST_ARCHLIB) -I\$(INST_LIB) -I\$(PERL_ARCHLIB) -I\$(PERL_LIB) t/5.close.t


   test6 : pure_all
\tPERL_DL_NONLAZY=1 \$(FULLPERL) -I\$(INST_ARCHLIB) -I\$(INST_LIB) -I\$(PERL_ARCHLIB) -I\$(PERL_LIB) t/6.delete.t

   test7 : pure_all
\tPERL_DL_NONLAZY=1 \$(FULLPERL) -I\$(INST_ARCHLIB) -I\$(INST_LIB) -I\$(PERL_ARCHLIB) -I\$(PERL_LIB) t/7.scan.t

   test8 : pure_all
\tPERL_DL_NONLAZY=1 \$(FULLPERL) -I\$(INST_ARCHLIB) -I\$(INST_LIB) -I\$(PERL_ARCHLIB) -I\$(PERL_LIB) t/8.sort.t

";
}
