/*
 * ZETA.xs - "the meat" of the entire package
 *
 * ZETA Perl - Perl binding for the ANSI/NISO Z39.50-1995 Protocol
 * Copyright (c) 1996, 1999 Finsiel S.p.A.
 *
 * Rocco Carbone <rocco@tecsiel.it> 4Q 1996
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef __cplusplus
extern "C" {
#endif
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#ifdef __cplusplus
}
#endif


/*
 * operating system header file(s)
 */
#if defined(_AIX)
#include <sys/select.h>
#endif

#if defined(WIN32)
/*
 * The following is to avoid header definition conflict between
 * the MSVC++ and Perl distribution for WIN32 platform. I had this
 * problem with MSVC++ 5.0 vs CPAN perl distribution (version 5.004_04
 * built for MSWin32) on my NT4.0 box.
 */
#undef select

#include "windows.h"
#include "winsock.h"

#define strcasecmp _stricmp

#endif


/*
 * ZETA header file(s)
 */
#include "zeta.h"

#define ZDEF_TIMEOUT  360

zattributesplusterm * newattributesplusterm (void);
zterm * newterm (char * string);
zoctetstring * zosalloc  (int size);
zquery * newquery (int querytype, zattributesetid * attrid);
char * search_error (zrecords * r);
char * external2a (zexternal * e);
char * buildrecords (zrecords * r);
zattributeelement * isknownattribute (char * a, zattributesetid * id);
char * tinfo2a (zentry * e);
char * buildentries (zlistentries * e);
char * builddiagnostics (unsigned long num, zdiagrec ** table);

zidauthentication * zauthidalloc (int choice);
void zauthidfree (zidauthentication * id);


/*
 * How a descriptor of a Z39.50 connection looks like
 */
typedef struct
{
  int fd;        /* unique identifier (Unix file descriptor)  */
  int state;     /* current state of the Z39.50 connection    */
  int event;     /* last event occurred                       */
  int versions;  /* protocol versions proposed/in-force       */
  int options;   /* protocol options proposed/in-force        */
  int msgsize;   /* preferred message size proposed/in-force  */
  int recsize;   /* exceptional record size proposed/in-force */
} zconnection;


/*
 * global variables here
 */
static zconnection * z = NULL;


/*
 * Prototypes
 */

char ** buildargv (char * s);
void freeargv (char ** s);



static int
not_here(s)
char *s;
{
    croak("%s not implemented on this architecture", s);
    return -1;
}

static double
constant(name, arg)
char *name;
int arg;
{
    errno = 0;
    return 0;

not_there:
    errno = ENOENT;
    return 0;
}


void
init_ZETA()
{
#if defined(ZAPI_TRACE)
  z_fd = fopen ("/tmp/zetagw.log", "a+");
  if (! z_fd)
    z_fd = stdout, z_trace = 0;
  else
    z_trace = Z_PDU_ONLY_BIT | Z_PID_BIT;
#endif /* ZAPI_TRACE */

  z = malloc (sizeof (zconnection));
  if (! z)
    croak ("Out of memory");

  z -> fd       = -1;
  z -> state    = 0;
  z -> event    = 0;
  z -> versions = Z_PROTOCOL_ALL;
  z -> options  = Z_ALLOWED_OPTIONS;
  z -> msgsize  = 32 * 1024;
  z -> recsize  = 40 * 1024;
}


MODULE = ZETA    PACKAGE = ZETA    PREFIX = zeta

REQUIRE: 1.935
PROTOTYPES: ENABLE

BOOT:
# first blank line terminates bootstrap code
init_ZETA ();

double
constant(name,arg)
	char *		name
	int		arg


 ##########################################################################
 # Low Level exportable routines
 ##########################################################################

 #
 # Disable tracing facilities.
 #
void
z_trace_off ()
   CODE:
{
#if defined(ZAPI_TRACE)
  if (z_fd != stdout)
    z_trace = Z_PDU_ONLY_BIT | Z_PID_BIT;
  else
    z_trace = 0;
#endif /* ZAPI_TRACE */
}


 #
 # Enable tracing facilities.
 #
void
z_trace_on ()
  CODE:
{
#if defined(ZAPI_TRACE)
  z_trace = Z_ALL_BITS;
#endif /* ZAPI_TRACE */
}


 #
 # Print software information.
 #
void
z_version ()
  CODE:
{
  z_version ();
}

 ##########################################################################
 # High Level exportable routines
 #
 # All calls are blocking/synchronous (i.e., a response must be received
 # or timeout before control is returned to the caller).
 ##########################################################################

MODULE = ZETA  PACKAGE = ZETA  PREFIX = zeta

 #
 # Try to connect at lower layer with a z39.50 Target
 #
zconnection *
zeta_connect (server = NULL, port = 2100)
        char * server
        int    port
CODE:
{
  int zresult;
  char target [BUFSIZ];
  char * p;

  RETVAL = z;

  /*
   * Establish a z39.50 connection.
   */
  z -> fd = z_open (NULL, Z_RDWR | Z_TCPIP_STACK);
  if (z -> fd != -1)
    {
      zresult = z_bind (z -> fd, "localhost", Z_ORIGIN);

      if (zresult != -1)
	{
	  /*
	   * Setup the Z39.50 Target Address
	   */
	  p = strchr (server, ':');
	  if (p)
	    port = atoi (p + 1);

	  sprintf (target, "%s:%d", server, port);

	  /*
	   * and try to connect now
	   */
	  zresult = z_connect (z -> fd, target);

	  if (zresult == -1)
	    {
	      z_close (z -> fd);
	      z -> fd = -1;
	      RETVAL = NULL;
	    }
	  else
	    {
	      RETVAL = z;
	    }
	}
      else
	{
	  z_close (z -> fd);
	  z -> fd = -1;
	  RETVAL = NULL;
	}
    }
  else
    {
      RETVAL = NULL;
    }
}
OUTPUT:
 RETVAL


 #
 # Send an Init Request Z39.50 APDU to Server
 # and wait for the Initialize Response.
 #
void
zeta_init (userid = NULL, passwd = NULL, refid = NULL, t_out = ZDEF_TIMEOUT)
	char * userid
	char * passwd
	char * refid
	int    t_out
PPCODE:
{
  /*
   * ZETA variables.
   */
  int ztype;
  int flags;
  zoctetstring * referenceid = NULL;                   /* Optional */

  /*
   * The Z39.50 APDUs to send and receive.
   */
  zpdu topdu;
  zpdu frompdu;
  zinitializerequest request;
  zinitializeresponse * response;
  /*
   * Authentication
   */
  zidauthentication * idauthentication = NULL;         /* Optional */

  /*
   * Reserve here space for the local variables.
   */
  int nready;
  fd_set readmask;
  struct timeval timeout;

  /*
   * The output variables to be pushed into the stack
   */
  int status = 0;              /* default to false */
  char buffer [200];
  SV * errtext = newSV (200);
  int result = 0;              /* result of the association */
  char format [10];
  SV * retref;
  /*
   * Default timeout to wait for the response.
   */
  int time_out = t_out;

  status = 0;
  sv_setpv (errtext, "");
  EXTEND (sp, 3);

  if (refid)
    {
      referenceid = zosalloc (strlen (refid));
      memcpy (referenceid -> data, refid, strlen (refid));
    }

  /*
   * Manage an optional userid and password
   */
  if (userid && * userid)
    {
      idauthentication = zauthidalloc (Z_AUTHTYPE_IDPASSWORD);
      idauthentication -> u . idpass -> userid = malloc (strlen (userid) + 1);
      strcpy (idauthentication -> u . idpass -> userid, userid);

      if (passwd && * passwd)
	{
	  idauthentication -> u . idpass -> password = malloc (strlen (passwd) + 1);
	  strcpy (idauthentication -> u . idpass -> password, passwd);
	}
    }


  /*
   * Fill the Z39.50 Init Request variable.
   */
  request . referenceid           = referenceid;
  request . protocolversion       = z -> versions = Z_PROTOCOL_ALL;
  request . options               = z -> options = Z_ALLOWED_OPTIONS;
  request . preferredmessagesize  = z -> msgsize = 32 * 1024;
  request . exceptionalrecordsize = z -> recsize = 40 * 1024;
  request . idauthentication      = idauthentication;;
  request . implementationid      = NULL;
  request . implementationname    = NULL;
  request . implementationversion = NULL;
  request . userinformationfield  = NULL;
  request . otherinfo             = NULL;

  topdu . u . initrequest = & request;

  if (z_snd (z -> fd, Z_INIT_REQ, & topdu, 0) == -1)
    {
      zosfree (referenceid);
      zauthidfree (idauthentication);
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "Cannot send INIT APDU [z_errno = %d - %s]",
	       z_errno, z_error ());
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (result)));
      return;
    }

  /*
   * call select(2) to perform synchronous I/O multiplexing.
   */
  FD_ZERO (& readmask);
  FD_SET (z -> fd, & readmask);

  timeout . tv_sec  = time_out;
  timeout . tv_usec = 0;

  zosfree (referenceid);
  zauthidfree (idauthentication);

  nready = select (z -> fd + 1, & readmask, (fd_set *) 0, (fd_set *) 0,
		   time_out > 0 ? & timeout : NULL);

  switch (nready)
    {
    case 0:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "No response in %d seconds", time_out);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (result)));
      break;

    case -1:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "select(2) failed - errno %d\n", errno);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (result)));
      break;

    default:
      /*
       * Receive the Z39.50 Init Response APDU.
       */
      if (z_rcv (z -> fd, & ztype, & frompdu, & flags) == -1)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Cannot receive INIT APDU [z_errno = %d - %s]",
		   z_errno, z_error ());
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (result)));
	}
      else if (ztype == Z_DISCON)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Disconnection by the network");
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (result)));
	}
      else if (ztype != Z_INIT_CNF)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer,
		   "Unexpected event %d received instead of expected %d\n",
		   ztype, Z_INIT_CNF);
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (result)));
	}
      else
	{
	  response = frompdu . u . initresponse;

	  switch (response -> result)
	    {
	    case Z_ACCEPTED:
	      status = 1;
	      result = 1;

	      PUSHs (sv_2mortal (newSViv (status)));
	      sv_setpv (errtext, "");
	      PUSHs (sv_2mortal (errtext));
	      PUSHs (sv_2mortal (newSViv (result)));
	      if (response -> referenceid && response -> referenceid -> data)
		refid = response -> referenceid -> data;
	      else
	        refid = NULL;
              EXTEND (sp, 1);
	      if (refid)
		{
		  retref = newSV (strlen (refid));
		  sv_setpv (retref, refid);
		  PUSHs (sv_2mortal (retref));
		}

	      z -> versions &= response -> protocolversion;
	      z -> options  &= response -> options;
	      z -> msgsize   = response -> preferredmessagesize;
	      z -> recsize   = response -> exceptionalrecordsize;

	      break;

	    default:
	      status = 0;
	      result = 0;

	      PUSHs (sv_2mortal (newSViv (status)));
	      sv_setpv (errtext, "rejected by target");
	      PUSHs (sv_2mortal (errtext));
	      PUSHs (sv_2mortal (newSViv (result)));
	      break;
	    }
	}
      break;
    }
}


 #
 # Send a Search Request Z39.50 APDU to Server
 # and wait for the Search Response.
 #
void
zeta_search (database, name, what, setid, t_out = ZDEF_TIMEOUT)
	char 	* database
	char 	* name
	char 	* what
	char    * setid
	int       t_out
PPCODE:
{
  extern zrpnquery * rpnquery;

  /*
   * ZETA variables.
   */
  int ztype;
  int flags;

  /*
   * The Z39.50 APDUs to send and receive.
   */
  zpdu topdu;
  zpdu frompdu;
  zsearchrequest request;
  zsearchresponse * response;

  /*
   * Reserve here space for the local variables.
   */
  int nready;
  fd_set readmask;
  struct timeval timeout;

  /*
   * The output variables to be pushed into the stack
   */
  int status = 0;              /* default to false */
  char buffer [200];
  SV * errtext = newSV (200);
  int found = 0;               /* # of records found */
  char * errbuff;
  /*
   * Default timeout to wait for the response.
   */
  int time_out = t_out;

  /*
   * Records size constraints.
   */
  int smallsetupperbound     = 0;                      /* Mandatory */
  int largesetlowerbound     = 1;                      /* Mandatory */
  int mediumsetpresentnumber = 0;                      /* Mandatory */
  /*
   * Replace indicator.
   */
  int replaceindicator = Z_TRUE;                       /* Mandatory */
  /*
   * Result set name.
   */
  zinternationalstring * resultsetname = NULL;         /* Mandatory */
  /*
   * Databases
   */
  zdatabasenames databases = {0, NULL};
  /*
   * Set elements names.
   */
  zelementsetnames brief = {Z_GENERIC, {Z_BRIEF}};
  zelementsetnames full  = {Z_GENERIC, {Z_FULL}};
  zelementsetnames * smallsetelementsetnames  = NULL;  /* Optional */
  zelementsetnames * mediumsetelementsetnames = NULL;  /* Optional */
  /*
   * Preferred record syntax.
   */
  zobjectid * preferredrecordsyntax = NULL;            /* Optional */
  /*
   * Query (only type-1 and type-101 are currently supported).
   */
  zattributesetid * attrid = NULL;                     /* Mandatory */
  zquery * query       = NULL;                         /* Mandatory */
  int querytype        = Z_QUERYTYPE_1;

  SV *info;
  /*
   * Add the database to the table of databases
   */
  addtotableofdatabases (& databases, database);

  /*
   * Get a query structure with null pointers for the term elements
   */
  if (setid && * setid)
    {
      if (! strcasecmp (setid, "bib1"))
	attrid = Z_BIB_1;
      else if (! strcasecmp (setid, "exp1"))
	attrid = Z_EXP_1;
      else if (! strcasecmp (setid, "ext1"))
	attrid = Z_EXT_1;
      else if (! strcasecmp (setid, "ccl1"))
	attrid = Z_CCL_1;
      else if (! strcasecmp (setid, "gils"))
	attrid = Z_GILS;
      else if (! strcasecmp (setid, "stas"))
	attrid = Z_STAS;
      else if (! strcasecmp (setid, "cimi"))
	attrid = Z_CIMI_1;
      else
	attrid = setid;            /* give a chance to the user too */
    }

  query = newquery (querytype, attrid);

  status = 0;
  sv_setpv (errtext, "");
  EXTEND (sp, 3);

  /*
   * Initialize pointer and parse the query
   */
  if (querytype == Z_QUERYTYPE_1)
    rpnquery = query -> u . type_1;
  else if (querytype == Z_QUERYTYPE_101)
    rpnquery = query -> u . type_101;
  else
    rpnquery = NULL;

  zeta_yyinit (what, strlen (what));
  if (zeta_yyparse ())
    {
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "Syntax error in \"%s\"", what);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (found)));
      return;
    }

  /*
   * Fill the Z39.50 Search Request variable.
   */
  request . referenceid              = NULL;
  request . smallsetupperbound       = smallsetupperbound;
  request . largesetlowerbound       = largesetlowerbound;
  request . mediumsetpresentnumber   = mediumsetpresentnumber;
  request . replaceindicator         = replaceindicator;
  request . resultsetname            = name ? name : Z_DEFAULT_RESULTSETNAME;
  request . databasenames            = & databases;
  request . smallsetelementsetnames  = smallsetelementsetnames;
  request . mediumsetelementsetnames = mediumsetelementsetnames;
  request . preferredrecordsyntax    = preferredrecordsyntax;
  request . query                    = query;
  request . additionalsearchinfo     = NULL;
  request . otherinfo                = NULL;

  topdu . u . searchrequest = & request;

  if (z_snd (z -> fd, Z_SEARCH_REQ, & topdu, 0) == -1)
    {
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "Cannot send SEARCH APDU [z_errno = %d - %s]",
	       z_errno, z_error ());
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (found)));
      return;
    }

  /*
   * call select(2) to perform synchronous I/O multiplexing.
   */
  FD_ZERO (& readmask);
  FD_SET (z -> fd, & readmask);

  timeout . tv_sec  = time_out;
  timeout . tv_usec = 0;

  nready = select (z -> fd + 1, & readmask, (fd_set *) 0, (fd_set *) 0,
		   time_out > 0 ? & timeout : NULL);

  switch (nready)
    {
    case 0:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "No response in %d seconds", time_out);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (found)));
      break;

    case -1:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "select(2) failed - errno %d\n", errno);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (found)));
      break;

    default:
      /*
       * Receive the Z39.50 Search Response APDU.
       */
      if (z_rcv (z -> fd, & ztype, & frompdu, & flags) == -1)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Cannot receive SEARCH APDU [z_errno = %d - %s]",
		   z_errno, z_error ());
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (found)));
	}
      else if (ztype == Z_DISCON)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Disconnection by the network");
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (found)));
	}
      else if (ztype != Z_SEARCH_CNF)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer,
		   "Unexpected event %d received instead of expected %d\n",
		   ztype, Z_SEARCH_CNF);
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (found)));
	}
      else
	{
	  response = frompdu . u . searchresponse;

	  found = response -> resultcount;
	  switch (response -> searchstatus)
	    {
	    case Z_TRUE:
	      status = 1;

	      PUSHs (sv_2mortal (newSViv (status)));
	      sv_setpv (errtext, "");
	      PUSHs (sv_2mortal (errtext));
	      PUSHs (sv_2mortal (newSViv (found)));

              if (response -> additionalsearchinfo &&
	          response -> additionalsearchinfo -> table &&
                  response ->additionalsearchinfo -> table[0] -> choice ==
	           Z_OTHERINFO_CHARACTERINFO &&
	           response -> additionalsearchinfo -> table[0] -> u .characterinfo)
	        {
	          info = newSV (strlen (response -> additionalsearchinfo ->
	                             table[0] -> u .characterinfo));
	          EXTEND (sp, 1);
	          sv_setpv (info, response -> additionalsearchinfo ->
	                    table[0] -> u .characterinfo);
	          PUSHs (sv_2mortal (info));

	        }
	      break;

	    default:
	      status = 0;

	      PUSHs (sv_2mortal (newSViv (status)));
	      errbuff = search_error(response -> records);
	      if (errbuff)
	        {
	          errtext = newSV(strlen(errbuff));
	          sv_setpv (errtext, errbuff);
	        }
	      else
		sv_setpv (errtext, "Search had a failure (unspecified)");

	      PUSHs (sv_2mortal (errtext));
	      PUSHs (sv_2mortal (newSViv (found)));
	      if (errbuff)
		free (errbuff);
	      break;
	    }
	}
      break;
    }
}


 #
 # Send a Present Request Z39.50 APDU to Server
 # and wait for the Present Response.
 #
void
zeta_present (name, howmany = 1, position = 1, esn = NULL, recsntx = NULL, t_out = ZDEF_TIMEOUT)
	char 	* name
	int	  howmany
	int	  position
	char    * esn
        char    * recsntx
	int       t_out
PPCODE:
{
  /*
   * ZETA variables.
   */
  int ztype;
  int flags;

  /*
   * The Z39.50 APDUs to send and receive.
   */
  zpdu topdu;
  zpdu frompdu;
  zpresentrequest request;
  zpresentresponse * response;

  /*
   * Reserve here space for the local variables.
   */
  int nready;
  fd_set readmask;
  struct timeval timeout;

  char ** argv = NULL;
  char ** q = NULL;
  int i;

  zobjectid * preferredrecordsyntax = NULL;

  /*
   * The output variables to be pushed into the stack
   */
  int status = 0;              /* default to false */
  char buffer [200];
  SV * errtext = newSV (200);
  int recordsno = 0;           /* # of records returned */
  int nextno = 0;              /* next result set position */
  char * records;              /* the records in a flat form */
  SV * found;

  /*
   * Default timeout to wait for the response.
   */
  int time_out = t_out;

  /*
   * Elements Set Names and Record Composition
   */
  zelementsetnames brief  = {Z_GENERIC, {Z_BRIEF}};
  zelementsetnames full   = {Z_GENERIC, {Z_FULL}};
  zelementsetnames * elementsetnames = NULL;           /* Optional */
  zrecordcomp composition = {Z_SIMPLE_RECORD_COMPOSITION, {NULL}};

  status = 0;
  sv_setpv (errtext, "");
  EXTEND (sp, 3);

  /*
   * Element Set Names.
   */
  if (esn && (! strcasecmp (esn, "brief") || ! strcasecmp (esn, "B")))
    elementsetnames = & brief;
  else if (esn && (! strcasecmp (esn, "full") || ! strcasecmp (esn, "F")))
    elementsetnames = & full;
  else
    elementsetnames = NULL;

  if (elementsetnames)
    composition . u . simple = elementsetnames;

  if (recsntx && * recsntx)
    {
      if (! strcmp (recsntx, "unimarc"))
	preferredrecordsyntax = Z_UNIMARC;
      else if (! strcmp (recsntx, "intermarc"))
	preferredrecordsyntax = Z_INTERMARC;
      else if (! strcmp (recsntx, "ccf"))
	preferredrecordsyntax = Z_CCF;
      else if (! strcmp (recsntx, "usmarc"))
	preferredrecordsyntax = Z_USMARC;
      else if (! strcmp (recsntx, "ukmarc"))
	preferredrecordsyntax = Z_UKMARC;
      else if (! strcmp (recsntx, "normarc"))
	preferredrecordsyntax = Z_NORMARC;
      else if (! strcmp (recsntx, "librismarc"))
	preferredrecordsyntax = Z_LIBRISMARC;
      else if (! strcmp (recsntx, "danmarc"))
	preferredrecordsyntax = Z_DANMARC;
      else if (! strcmp (recsntx, "finmarc"))
	preferredrecordsyntax = Z_FINMARC;
      else if (! strcmp (recsntx, "mab"))
	preferredrecordsyntax = Z_MAB;
      else if (! strcmp (recsntx, "canmarc"))
	preferredrecordsyntax = Z_CANMARC;
      else if (! strcmp (recsntx, "sbn"))
	preferredrecordsyntax = Z_SBN;
      else if (! strcmp (recsntx, "picamarc"))
	preferredrecordsyntax = Z_PICAMARC;
      else if (! strcmp (recsntx, "AUSMARC"))
	preferredrecordsyntax = Z_AUSMARC;
      else if (! strcmp (recsntx, "ibermarc"))
	preferredrecordsyntax = Z_IBERMARC;
      else if (! strcmp (recsntx, "explain"))
	preferredrecordsyntax = Z_EXPLAIN;
      else if (! strcmp (recsntx, "sutrs"))
	preferredrecordsyntax = Z_SUTRS;
      else if (! strcmp (recsntx, "opac"))
	preferredrecordsyntax = Z_OPAC;
      else if (! strcmp (recsntx, "summary"))
	preferredrecordsyntax = Z_SUMMARY;
      else if (! strcmp (recsntx, "grs1"))
	preferredrecordsyntax = Z_GRS_1;
      else if (! strcmp (recsntx, "extendedservices"))
	preferredrecordsyntax = Z_EXTENDED_SERVICES;
      else
	preferredrecordsyntax = recsntx;  /* give a chance to the user too */
    }

  /*
   * Fill the Z39.50 Present Request variable.
   */
  request . referenceid              = NULL;
  request . resultsetid              = name ? name : Z_DEFAULT_RESULTSETNAME;
  request . resultsetstartpoint      = position;
  request . numberofrecordsrequested = howmany;
  request . additionalranges         = NULL;
  request . recordcomposition        = elementsetnames ? & composition : NULL;
  request . preferredrecordsyntax    = preferredrecordsyntax;
  request . maxsegmentcount          = NULL;
  request . maxrecordsize            = NULL;
  request . maxsegmentsize           = NULL;
  request . otherinfo                = NULL;

  topdu . u . presentrequest = & request;

  if (z_snd (z -> fd, Z_PRESENT_REQ, & topdu, 0) == -1)
    {
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "Cannot send PRESENT APDU [z_errno = %d - %s]",
	       z_errno, z_error ());
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (recordsno)));
      return;
    }

  /*
   * call select(2) to perform synchronous I/O multiplexing.
   */
  FD_ZERO (& readmask);
  FD_SET (z -> fd, & readmask);

  timeout . tv_sec  = time_out;
  timeout . tv_usec = 0;

  nready = select (z -> fd + 1, & readmask, (fd_set *) 0, (fd_set *) 0,
		   time_out > 0 ? & timeout : NULL);

  switch (nready)
    {
    case 0:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "No response in %d seconds", time_out);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (recordsno)));
      break;

    case -1:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "select(2) failed - errno %d\n", errno);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (recordsno)));
      break;

    default:
      /*
       * Receive the Z39.50 Present Response APDU.
       */
      if (z_rcv (z -> fd, & ztype, & frompdu, & flags) == -1)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Cannot receive PRESENT APDU [z_errno = %d - %s]",
		   z_errno, z_error ());
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (recordsno)));
	}
      else if (ztype == Z_DISCON)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Disconnection by the network");
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (recordsno)));
	}
      else if (ztype != Z_PRESENT_CNF)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer,
		   "Unexpected event %d received instead of expected %d\n",
		   ztype, Z_PRESENT_CNF);
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (recordsno)));
	}
      else
	{
	  response = frompdu . u . presentresponse;

	  recordsno = response -> numberofrecordsreturned;
	  nextno    = response -> nextresultsetposition;

	  switch (response -> presentstatus)
	    {
	    case Z_SUCCESS_STATUS:
	    case Z_PARTIAL_2_STATUS:

	      status = 1;

	      PUSHs (sv_2mortal (newSViv (status)));
	      sv_setpv (errtext, "");
	      PUSHs (sv_2mortal (errtext));
	      PUSHs (sv_2mortal (newSViv (recordsno)));

	      if (response -> records &&
		  response -> records -> choice == Z_RESPONSERECORDS_RECORDS &&
		  response -> records -> u . responserecords &&
		  response -> records -> u . responserecords -> table)
		{
		  znameplusrecord * npr;

		  for (i = 0;
		       i < response -> records -> u . responserecords -> size;
		       i ++)
		    {
		      npr = response -> records -> u . responserecords -> table [i];

		      if (npr && npr -> record &&
			  npr -> record -> choice == Z_RETRIEVAL_RECORD &&
			  npr -> record -> u . retrievalrecord)
			{
			  records =
			    external2a (npr -> record -> u . retrievalrecord);

			  EXTEND (sp, 1);
			  found = newSV (strlen (records));
			  sv_setpv (found, records);
			  PUSHs (sv_2mortal (found));
			}
		    }
		}
	      break;

	    default:
	      status = 0;

	      PUSHs (sv_2mortal (newSViv (status)));

	      /*
	       * Get a buffer with all the non surrogate diagnostics.
	       */
	      records = buildrecords (response -> records);
	      if (records)
		{
		  errtext = newSV (strlen (records));
		  sv_setpv (errtext, records);
		}
	      else
		sv_setpv (errtext, "Cannot build reason");
	      PUSHs (sv_2mortal (errtext));

	      PUSHs (sv_2mortal (newSViv (recordsno)));

	      if (records)
		free (records);
	      break;
	    }
	}
      break;
    }
}


 #
 # Send a Delete Request Z39.50 APDU to Server
 # and wait for the Delete Response.
 #
void
zeta_delete (function, name, t_out = ZDEF_TIMEOUT)
	int 	  function
	char 	* name
	int       t_out
PPCODE:
{
  /*
   * ZETA variables.
   */
  int ztype;
  int flags;

  /*
   * The Z39.50 APDUs to send and receive.
   */
  zpdu topdu;
  zpdu frompdu;
  zdeleterequest request;
  zdeleteresponse * response;

  /*
   * Table of result set.
   */
  zresultsetlist rstable = {0, NULL};                  /* Mandatory */

  /*
   * Reserve here space for the local variables.
   */
  int nready;
  fd_set readmask;
  struct timeval timeout;

  /*
   * The output variables to be pushed into the stack
   */
  int status = 0;              /* default to false */
  char buffer [200];
  SV * errtext = newSV (200);

  /*
   * Default timeout to wait for the response.
   */
  int time_out = t_out;

  status = 0;
  sv_setpv (errtext, "");
  EXTEND (sp, 2);

  /*
   * Add the resultset to the table of resultsets
   */
  addtotableofresultset (& rstable, name);

  /*
   * Fill the Z39.50 Delete Request variable.
   */
  request . referenceid   = NULL;
  request . function      = function;
  request . resultsetlist = & rstable;
  request . otherinfo     = NULL;

  topdu . u . deleterequest = & request;

  if (z_snd (z -> fd, Z_DELETE_REQ, & topdu, 0) == -1)
    {
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "Cannot send DELETE APDU [z_errno = %d - %s]",
	       z_errno, z_error ());
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      return;
    }

  /*
   * call select(2) to perform synchronous I/O multiplexing.
   */
  FD_ZERO (& readmask);
  FD_SET (z -> fd, & readmask);

  timeout . tv_sec  = time_out;
  timeout . tv_usec = 0;

  nready = select (z -> fd + 1, & readmask, (fd_set *) 0, (fd_set *) 0,
		   time_out > 0 ? & timeout : NULL);

  switch (nready)
    {
    case 0:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "No response in %d seconds", time_out);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      break;

    case -1:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "select(2) failed - errno %d\n", errno);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      break;

    default:
      /*
       * Receive the Z39.50 Delete Response APDU.
       */
      if (z_rcv (z -> fd, & ztype, & frompdu, & flags) == -1)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Cannot receive DELETE APDU [z_errno = %d - %s]",
		   z_errno, z_error ());
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else if (ztype == Z_DISCON)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Disconnection by the network");
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else if (ztype != Z_DELETE_CNF)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer,
		   "Unexpected event %d received instead of expected %d\n",
		   ztype, Z_DELETE_CNF);
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else
	{
	  status = 1;

	  PUSHs (sv_2mortal (newSViv (status)));
	  sv_setpv (errtext, "");
	  PUSHs (sv_2mortal (errtext));

	  response = frompdu . u . deleteresponse;
	}
      break;
    }
}


 #
 # Send a Scan Request Z39.50 APDU to Server
 # and wait for the Scan Response.
 #
void
zeta_scan (database, term, howmany = 1, preferred = -1, t_out = ZDEF_TIMEOUT)
	char 	* database
	char 	* term
	int	howmany
	int	preferred
	int     t_out
PPCODE:
{
  extern zrpnquery * rpnquery;

  /*
   * ZETA variables.
   */
  int ztype;
  int flags;

  /*
   * The Z39.50 APDUs to send and receive.
   */
  zpdu topdu;
  zpdu frompdu;
  zscanrequest request;
  zscanresponse * response;

  /*
   * Reserve here space for the local variables.
   */
  int nready;
  fd_set readmask;
  struct timeval timeout;

  char ** argv;
  char ** q = NULL;
  int i;
  char * current = NULL;

  /*
   * The output variables to be pushed into the stack
   */
  int status = 0;              /* default to false */
  char buffer [200];
  SV * errtext = newSV (200);
  int entriesno = 0;           /* # of entries returned */
  char * entries;              /* the entries in a flat form */
  SV * found;

  /*
   * Default timeout to wait for the response.
   */
  int time_out = t_out;

  /*
   * Databases
   */
  zdatabasenames databases = {0, NULL};
  /*
   * Window size constraints.
   */
  int stepsize  = -1;                                  /* Optional */
  int requested = 1;                                   /* Mandatory */
  int position  = preferred;                           /* Optional */
  /*
   * Term
   */
  zattributeelement * elem = NULL;
  zattributesplusterm * apt = newattributesplusterm ();

  /*
   * Add the database to the table of databases
   */
  addtotableofdatabases (& databases, database);

  status = 0;
  sv_setpv (errtext, "");
  EXTEND (sp, 3);

  /*
   * Process the term and fill in the attributesplusterm structure
   */

  argv = buildargv (term);
  for (q = argv; q && (* q) && (* q [0]); q ++)
    {
      /*
       * Process first an optional attribute
       */
      elem = isknownattribute (* q, NULL);
      if (elem)
	addtolistofattr (apt -> attributes, elem);
      else
	{
	  if (! apt -> term)
	    /*
	     * Process now the term (it should be mandatory)
	     */
	    apt -> term = newterm (* q);
	  else
	    {
	      /*
	       * Too many terms. Drop it (silently)
	       */
	      ;
	    }
	}
    }
  if (argv)
    freeargv (argv);


  /*
   * Fill the Z39.50 Scan Request variable.
   */
  request . referenceid                 = NULL;
  request . databasenames               = & databases;
  request . attributeset                = NULL;
  request . termlistandstartpoint       = apt;
  request . stepsize                    = NULL;
  request . numberoftermsrequested      = howmany;
  request . preferredpositioninresponse =
    position == -1 ? NULL : & position;
  request . otherinfo                   = NULL;

  topdu . u . scanrequest = & request;

  if (z_snd (z -> fd, Z_SCAN_REQ, & topdu, 0) == -1)
    {
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "Cannot send SCAN APDU [z_errno = %d - %s]",
	       z_errno, z_error ());
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (entriesno)));
      return;
    }

  /*
   * call select(2) to perform synchronous I/O multiplexing.
   */
  FD_ZERO (& readmask);
  FD_SET (z -> fd, & readmask);

  timeout . tv_sec  = time_out;
  timeout . tv_usec = 0;

  nready = select (z -> fd + 1, & readmask, (fd_set *) 0, (fd_set *) 0,
		   time_out > 0 ? & timeout : NULL);

  switch (nready)
    {
    case 0:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "No response in %d seconds", time_out);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (entriesno)));
      break;

    case -1:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "select(2) failed - errno %d\n", errno);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      PUSHs (sv_2mortal (newSViv (entriesno)));
      break;

    default:
      /*
       * Receive the Z39.50 Scan Response APDU.
       */
      if (z_rcv (z -> fd, & ztype, & frompdu, & flags) == -1)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Cannot receive SCAN APDU [z_errno = %d - %s]",
		   z_errno, z_error ());
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (entriesno)));
	}
      else if (ztype == Z_DISCON)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Disconnection by the network");
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (entriesno)));
	}
      else if (ztype != Z_SCAN_CNF)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer,
		   "Unexpected event %d received instead of expected %d\n",
		   ztype, Z_SCAN_CNF);
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	  PUSHs (sv_2mortal (newSViv (entriesno)));
	}
      else
	{
	  response = frompdu . u . scanresponse;

	  entriesno = response -> numberofentriesreturned;

	  switch (response -> scanstatus)
	    {
	    case Z_SUCCESS_STATUS:
	    case Z_PARTIAL_2_STATUS:
	    case Z_PARTIAL_5_STATUS:
	      status = 1;

	      PUSHs (sv_2mortal (newSViv (status)));
	      sv_setpv (errtext, "");
	      PUSHs (sv_2mortal (errtext));
	      PUSHs (sv_2mortal (newSViv (entriesno)));

	      if (response -> entries && response -> entries -> entries)
		{
		  for (i = 0; i < response -> entries -> entries -> size; i ++)
		    {
		      current =
			tinfo2a (response -> entries -> entries -> table [i]);
		      if (current)
			{
			  EXTEND (sp, 1);
			  found = newSV (strlen (current));
			  sv_setpv (found, current);
			  PUSHs (sv_2mortal (found));
			}
		    }
		}
	      break;

	    default:
	      status = 0;

	      PUSHs (sv_2mortal (newSViv (status)));

	      /*
	       * Get a buffer with all the non surrogate diagnostics.
	       */
	      entries = buildentries (response -> entries);

	      errtext = newSV (strlen (entries));
	      sv_setpv (errtext, entries);
	      PUSHs (sv_2mortal (errtext));

	      PUSHs (sv_2mortal (newSViv (entriesno)));

	      if (entries)
		free (entries);
	      break;
	    }
	}
      break;
    }
}


 #
 # Send a Sort Request Z39.50 APDU to Server
 # and wait for the Sort Response.
 #
void
zeta_sort (name, criteria, t_out = ZDEF_TIMEOUT)
	char 	* name
	char 	* criteria
	int       t_out
PPCODE:
{
  /*
   * ZETA variables.
   */
  int ztype;
  int flags;

  /*
   * The Z39.50 APDUs to send and receive.
   */
  zpdu topdu;
  zpdu frompdu;
  zsortrequest request;
  zsortresponse * response;
  char *diags;
  /*
   * Reserve here space for the local variables.
   */
  int nready;
  fd_set readmask;
  struct timeval timeout;

  /*
   * The output variables to be pushed into the stack
   */
  int status = 0;              /* default to false */
  char buffer [200];
  SV * errtext = newSV (200);
  int found = 0;               /* # of records found */
  int type;
  int value;
  char * s;
  char * e;
  int len;

  /*
   * Default timeout to wait for the response.
   */
   int time_out = t_out;

  /*
   * Sort Relation
   */
  int relation    = Z_ASCENDING_SORTRELATION;
  int sensitivity = Z_CASEINSENSITIVE;

  /*
   * Sorted result set name
   */
  zinternationalstring * sortname = name;

  /*
   * Table of input result set
   */
  zinputresultsetnames rsname = {0, NULL};

  /*
   * Sort Sequence
   */
  zattributesetid * id = Z_BIB_1;
  zsortsequence sequence = {0, NULL};


  /*
   * Add the name to the table of input result sets
   */
  addtoinputresultset (& rsname, name);

  /*
   * Process the element and fill in the sort sequence structure
   */
  s = e = criteria;
  len = strlen (criteria);

  while (* s == ' ')
    s ++;

  while (* s && len > 0  && sscanf (s, "%d=%d", & type, & value) == 2)
    {
      e = s;

      /*
       * look for the first blank or null
       */
      while (* e && * e != ' ')
	e ++, len --;
      * e = '\0';

      addtokeyspec (& sequence, s, id, relation, sensitivity);

      if (len)
	e ++;

      while (len > 0 && * e == ' ')
	e ++, len --;

      s = e;
    }

  status = 0;
  sv_setpv (errtext, "");
  EXTEND (sp, 2);

  /*
   * Fill the Z39.50 Sort Request variable.
   */
  request . referenceid              = NULL;
  request . inputresultsetnames      = & rsname;
  request . sortedresultsetname      = name;
  request . sortsequence             = & sequence;
  request . otherinfo                = NULL;

  topdu . u . sortrequest = & request;

  if (z_snd (z -> fd, Z_SORT_REQ, & topdu, 0) == -1)
    {
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "Cannot send SORT APDU [z_errno = %d - %s]",
	       z_errno, z_error ());
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      return;
    }

  /*
   * call select(2) to perform synchronous I/O multiplexing.
   */
  FD_ZERO (& readmask);
  FD_SET (z -> fd, & readmask);

  timeout . tv_sec  = time_out;
  timeout . tv_usec = 0;

  nready = select (z -> fd + 1, & readmask, (fd_set *) 0, (fd_set *) 0,
		   time_out > 0 ? & timeout : NULL);

  switch (nready)
    {
    case 0:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "No response in %d seconds", time_out);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      break;

    case -1:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "select(2) failed - errno %d\n", errno);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      break;

    default:
      /*
       * Receive the Z39.50 Sort Response APDU.
       */
      if (z_rcv (z -> fd, & ztype, & frompdu, & flags) == -1)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Cannot receive SORT APDU [z_errno = %d - %s]",
		   z_errno, z_error ());
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else if (ztype == Z_DISCON)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Disconnection by the network");
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else if (ztype != Z_SORT_CNF)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer,
		   "Unexpected event %d received instead of expected %d\n",
		   ztype, Z_SORT_CNF);
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else
	{
	  response = frompdu . u . sortresponse;

	  switch (response -> sortstatus)
	    {
	    case Z_TRUE:
	      status = 1;

	      PUSHs (sv_2mortal (newSViv (status)));
	      sv_setpv (errtext, "");
	      PUSHs (sv_2mortal (errtext));

	      break;

	    default:
	      status = 0;

	      PUSHs (sv_2mortal (newSViv (status)));
	      /*
	       * Get a buffer with all the surrogate diagnostics.
	       */
	      diags = builddiagnostics (response -> diagnostics -> size,
		                        response -> diagnostics -> table);

	      errtext = newSV (strlen (diags));
	      sv_setpv (errtext, diags);
	      PUSHs (sv_2mortal (errtext));
	      if (diags)
		free (diags);
	      break;
	    }
	}
      break;
    }
}


 #
 # Send a Close Request Z39.50 APDU to Server
 # and wait for the Close Response.
 #
void
zeta_close (reason, t_out = ZDEF_TIMEOUT)
	int reason
	int t_out
PPCODE:
{
  /*
   * ZETA variables.
   */
  int ztype;
  int flags;

  /*
   * The Z39.50 APDUs to send and receive.
   */
  zpdu topdu;
  zpdu frompdu;
  zclose request;
  zclose * response;

  /*
   * Reserve here space for the local variables.
   */
  int nready;
  fd_set readmask;
  struct timeval timeout;

  /*
   * The output variables to be pushed into the stack
   */
  int status = 0;              /* default to false */
  char buffer [200];
  SV * errtext = newSV (200);

  /*
   * Default timeout to wait for the response.
   */
  int time_out = t_out;

  status = 0;
  sv_setpv (errtext, "");
  EXTEND (sp, 2);

  /*
   * Fill the Z39.50 Close Request variable.
   */
  request . referenceid          = NULL;
  request . reason               = reason;
  request . diagnostic           = NULL;
  request . resourcereportformat = NULL;
  request . resourcereport       = NULL;
  request . otherinfo            = NULL;


  topdu . u . close = & request;

  if (z_snd (z -> fd, Z_CLOSE_REQ, & topdu, 0) == -1)
    {
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "Cannot send CLOSE APDU [z_errno = %d - %s]",
	       z_errno, z_error ());
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      return;
    }

  /*
   * call select(2) to perform synchronous I/O multiplexing.
   */
  FD_ZERO (& readmask);
  FD_SET (z -> fd, & readmask);

  timeout . tv_sec  = time_out;
  timeout . tv_usec = 0;

  nready = select (z -> fd + 1, & readmask, (fd_set *) 0, (fd_set *) 0,
		   time_out > 0 ? & timeout : NULL);

  switch (nready)
    {
    case 0:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "No response in %d seconds", time_out);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      break;

    case -1:
      PUSHs (sv_2mortal (newSViv (status)));
      sprintf (buffer, "select(2) failed - errno %d\n", errno);
      sv_setpv (errtext, buffer);
      PUSHs (sv_2mortal (errtext));
      break;

    default:
      /*
       * Receive the Z39.50 Close Response APDU.
       */
      if (z_rcv (z -> fd, & ztype, & frompdu, & flags) == -1)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Cannot receive CLOSE APDU [z_errno = %d - %s]",
		   z_errno, z_error ());
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else if (ztype == Z_DISCON)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer, "Disconnection by the network");
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else if (ztype != Z_CLOSE_CNF)
	{
	  PUSHs (sv_2mortal (newSViv (status)));
	  sprintf (buffer,
		   "Unexpected event %d received instead of expected %d\n",
		   ztype, Z_CLOSE_CNF);
	  sv_setpv (errtext, buffer);
	  PUSHs (sv_2mortal (errtext));
	}
      else
	{
	  response = frompdu . u . close;

	  switch (response -> reason)
	    {
	    case Z_FINISHED:
	    case Z_SHUTDOWN:
	    case Z_SYSTEM_PROBLEM:
	    case Z_COST_LIMIT:
	    case Z_RESOURCES:
	    case Z_SECURITY_VIOLATION:
	    case Z_PROTOCOL_ERROR:
	    case Z_LACK_OF_ACTIVITY:
	    case Z_PEER_ABORT:
	    case Z_UNSPECIFIED:
	      status = 1;

	      PUSHs (sv_2mortal (newSViv (status)));
	      sv_setpv (errtext, "");
	      PUSHs (sv_2mortal (errtext));

	      break;

	    default:
	      status = 0;

	      PUSHs (sv_2mortal (newSViv (status)));
	      sv_setpv (errtext, "Impossible reason");
	      PUSHs (sv_2mortal (errtext));
	      break;
	    }
	}
      break;
    }
}
