/*
 *                                                                   -*- C -*-
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * grs1.h - definitions needed to ANSI/NISO Z39.50-1995 users
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 *          #######    ##                      ##            ####
 *          ##         ##                      ##              ##
 *          ##                                                 ##
 *          ##       ####   ######   #####   ####    #####     ##
 *          #####      ##   ##   ## ##   ##    ##   ##   ##    ##
 *          ##         ##   ##   ## ##         ##   ##   ##    ##
 *          ##         ##   ##   ##  #####     ##   #######    ##
 *          ##         ##   ##   ##      ##    ##   ##         ##
 *          ##         ##   ##   ## ##   ##    ##   ##         ##
 *          ##       ###### ##   ##  #####   ######  ######  ######
 *                                                                  S.p.A.
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * Product Name  : ZETA - ANSI/NISO Z39.50-1995 services for Un*x Systems
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * Copyright (c) : 1996, 1999 Finsiel S.p.A. - All Rights Reserved
 * Finsiel Lab   : 34/B, Via Matteucci I-56124 Pisa - Italy
 * Author        : Rocco Carbone <rocco@tecsiel.it>
 * Created       : 1Q 97
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 *  THE SOFTWARE PROGRAMS CONTAINED OR DESCRIBED HEREIN ARE CONFIDENTIAL
 * INFORMATION AND PROPRIETARY PRODUCTS OF FINSIEL S.p.A. OR ITS LICENSEE
 *
 *                               NOTICE
 *
 * Use, copy, modification and distribution of this software is subject
 * to restrictions of valid written license from Finsiel S.p.A.
 *
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 */

typedef struct genericrecord genericrecord;


#define GRS1_OCTETS            1
#define GRS1_NUMERIC           2
#define GRS1_DATE              3
#define GRS1_EXT               4
#define GRS1_STRING            5
#define GRS1_TRUEORFALSE       6
#define GRS1_OID               7
#define GRS1_INTUNIT           8
#define GRS1_ELEMENTNOTTHERE   9
#define GRS1_ELEMENTEMPTY     10
#define GRS1_NODATAREQUESTED  11
#define GRS1_DIAGNOSTIC       12
#define GRS1_SUBTREE          13
typedef struct
{
  int choice;
  union
  {
    zoctetstring         * octet;
    int                    numeric;
    zgeneralizedtime     * date;
    zexternal            * external;
    zinternationalstring * string;
    int                    trueorfalse;
    zobjectid            * oid;
    zintunit             * intunit;
    char                   elementnotthere;
    char                   elementempty;
    char                   nodatarequested;
    zexternal            * diagnostic;
    genericrecord        * subtree;
  } u;
} grs1_data;


typedef struct
{
  int ascending;
  int order;
} grs1_order;


#define GRS1_USAGE_REDISTRIBUTABLE 1
#define GRS1_USAGE_RESTRICTED      2
#define GRS1_USAGE_LICENSEPOINTER  3
typedef struct
{
  int type;
  zinternationalstring * restriction;        /* OPTIONAL */
} grs1_usage;


typedef struct
{
  int              * type;                   /* OPTIONAL */
  zstringornumeric * value;
  int              * occurrence;             /* OPTIONAL */
} grs1_tagpath;


typedef struct
{
  int             size;
  grs1_tagpath ** table;           /* pointer to the table of pointers */
} grs1_tagpaths;


typedef struct
{
  zterm        * satisfier;     /* OPTIONAL */
  zintunit     * offset;        /* OPTIONAL */
  zintunit     * length;        /* OPTIONAL */
  int          * rank;          /* OPTIONAL */
  zoctetstring * token;         /* OPTIONAL */
} grs1_hitvector;


typedef struct
{
  int               size;
  grs1_hitvector ** table;       /* pointer to the table of pointers */
} grs1_hits;


#define Z_TRIPLE_INTEGER             1
#define Z_TRIPLE_INTERNATIONALSTRING 2
#define Z_TRIPLE_OCTETSTRING         3
#define Z_TRIPLE_OID                 4
#define Z_TRIPLE_BOOLEAN             5
#define Z_TRIPLE_NULL                6
#define Z_TRIPLE_UNIT                7
#define Z_TRIPLE_VALUEANDUNIT        8

typedef struct
{
  zobjectid * setid;            /* OPTIONAL */
  int         class;
  int         type;
  int         choice;
  union
  {
    int                    integer;
    zinternationalstring * internationalstring;
    zoctetstring         * octetstring;
    zobjectid            * oid;
    int                    boolean;
    char                   null;
    zunit                * unit;
    zintunit             * valueandunit;
  } value;
} grs1_triple;



typedef struct
{
  int            size;
  grs1_triple ** table;           /* pointer to the table of pointers */
} grs1_triples;


typedef struct
{
  zobjectid     * globalsetid;               /* OPTIONAL */
  grs1_triples  * triples;
} grs1_variant;


typedef struct
{
  int            size;
  grs1_variant ** table;          /* pointer to the table of pointers */
} grs1_variants;


typedef struct
{
  grs1_order           * seriesorder;         /* OPTIONAL */
  grs1_usage           * usageright;          /* OPTIONAL */
  grs1_hits            * hits;                /* OPTIONAL */
  zinternationalstring * displayname;         /* OPTIONAL */
  grs1_variants        * supportedvariants;   /* OPTIONAL */
  zinternationalstring * message;             /* OPTIONAL */
  zoctetstring         * elementdescriptor;   /* OPTIONAL */
  grs1_tagpaths        * surrogatefor;        /* OPTIONAL */
  grs1_tagpaths        * surrogateelement;    /* OPTIONAL */
  zexternal            * other;               /* OPTIONAL */
} grs1_metadata;


typedef struct
{
  int              * type;              /* OPTIONAL */
  zstringornumeric * value;
  int              * occurrence;        /* OPTIONAL */
  grs1_data        * content;
  grs1_metadata    * metadata;          /* OPTIONAL */
  grs1_variant     * appliedvariant;    /* OPTIONAL */
} grs1_taggedelement;



/*
 * A Generic Record is represented by a simple tree or a sequence of trees.
 *
 * An abstract database record could be represented by a single
 * instance of just one tagged element (the root of a single tree) or multiple
 * instances of tagged elements in a flat sequence (multiple trees)
 */
typedef struct
{
  int                   size;
  grs1_taggedelement ** tree;
} grs1_genericrecord;



DECEXP grs1_genericrecord * EXPORT_FUNC maketree (grs1_genericrecord * g, char * f, char * val);
DECEXP grs1_genericrecord * EXPORT_FUNC maketree_rep (grs1_genericrecord * g, char * format);
DECEXP void    EXPORT_FUNC grs1_free (grs1_genericrecord * g);
DECEXP grs1_genericrecord *  EXPORT_FUNC del_grs1_subtree (grs1_genericrecord * g);
DECEXP zoctetstring *  EXPORT_FUNC grs1ber (grs1_genericrecord * grs1, zoctetstring * os);
