/*
 *                                                                   -*- C -*-
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * zeta.h - definitions needed to ANSI/NISO Z39.50-1995 users
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 *          #######    ##                      ##            ####
 *          ##         ##                      ##              ##
 *          ##                                                 ##
 *          ##       ####   ######   #####   ####    #####     ##
 *          #####      ##   ##   ## ##   ##    ##   ##   ##    ##
 *          ##         ##   ##   ## ##         ##   ##   ##    ##
 *          ##         ##   ##   ##  #####     ##   #######    ##
 *          ##         ##   ##   ##      ##    ##   ##         ##
 *          ##         ##   ##   ## ##   ##    ##   ##         ##
 *          ##       ###### ##   ##  #####   ######  ######  ######
 *                                                                  S.p.A.
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * Product Name  : ZETA - ANSI/NISO Z39.50-1995 services for Un*x Systems
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * Copyright (c) : 1996, 1999 Finsiel S.p.A. - All Rights Reserved
 * Finsiel Lab   : 34/B, Via Matteucci I-56124 Pisa - Italy
 * Author        : Rocco Carbone <rocco@tecsiel.it>
 * Created       : 2Q 96
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 *  THE SOFTWARE PROGRAMS CONTAINED OR DESCRIBED HEREIN ARE CONFIDENTIAL
 * INFORMATION AND PROPRIETARY PRODUCTS OF FINSIEL S.p.A. OR ITS LICENSEE
 *
 *                               NOTICE
 *
 * Use, copy, modification and distribution of this software is subject
 * to restrictions of valid written license from Finsiel S.p.A.
 *
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 */


/*
 * Constants and data structures definitions for ANSI/NISO Z39.50-1995 users
 */



/*
 * Error variables
 */
extern int    z_errno;
extern char * z_errlist [];



/*
 * Error codes for the ANSI/NISO Z39.50-1995 Library Interface.
 */

#define Z_NOERROR              (0) /* No ploblem (yet !)                 */
#define Z_NOTSUPPORTED         (1) /* API function not yet supported     */
#define Z_BADNAME              (2) /* Invalid protocol name              */
#define Z_NOSTACK              (3) /* Communication stack not available  */
#define Z_BADFLAGS             (4) /* Wrong flags                        */
#define Z_ENOMEM               (5) /* Not enough memory                  */
#define Z_NOFD                 (6) /* An enpoint cannot be created       */
#define Z_NOASYNCMODE          (7) /* Asynchronous mode cannot be set    */
#define Z_BADF                 (8) /* Illegal Z39.50 endpoint identifier */
#define Z_OUTSTATE             (9) /* Function was issued in wrong state */
#define Z_BADROLE             (10) /* Wrong role                         */
#define Z_BADADDR             (11) /* Address was in an incorrect format */
#define Z_HOSTNOTFOUND        (12) /* No such host is known              */
#define Z_NODATA              (13) /* No data available                  */
#define Z_ECONNREFUSED        (14) /* Connection refused                 */
#define Z_BUFOVFLW            (15) /* Buffer not large enough            */
#define Z_SYSERR              (16) /* System error occurred              */
#define Z_SFIGA               (17) /* This error cannot be named         */
#define Z_BADPRIM             (18) /* No such primitive                  */
#define Z_PRIMNOTSUPPORTED    (19) /* Primitive not supported (yet)      */
#define Z_BADPARAMETER        (20) /* Wrong parameters                   */
#define Z_BADVERSION          (21) /* Wrong protocol version             */
#define Z_BADOPTION           (22) /* Wrong (or unsupported) options     */
#define Z_BADAUTHID           (23) /* Wrong authorization identifier     */
#define Z_BADBOUNDS           (24) /* Illegal result set bounds          */
#define Z_BADREPLACE          (25) /* Illegal replace indicator          */
#define Z_BADDATABASE         (26) /* Missing or illegal database        */
#define Z_QUERYNOTSUPPORTED   (27) /* Query type not supported (yet)     */
#define Z_BADQUERYTYPE        (28) /* Illegal query type                 */
#define Z_BADQUERY            (29) /* Illegal query                      */
#define Z_ENCODING            (30) /* BER coding error                   */
#define Z_DECODING            (31) /* BER decoding error                 */
#define Z_EPROTO              (32) /* Protocol error                     */
#define Z_SOFTWARE            (33) /* Internal software error            */
#define Z_UNAVAILABLE         (34) /* No such capability is in effect    */

#define Z_NERR                (34)



/*
 * Constants for `oflag' to be used with the z_open()
 * including definition of supported stacks and execution mode.
 */
#define Z_TCPIP_STACK  (0x0001)       /* RFC1729 and TCP/IP stack */
#define Z_XMOSI_STACK  (0x0002)       /* Finsiel X/mOSI stack */
#define Z_RDWR         (0x0004)       /* Open for reading or writing */
#define Z_NONBLOCK     (0x0008)       /* Non-blocking open and/or I/O */
#define Z_NDELAY       Z_NONBLOCK     /* only for backward compatibility */

/*
 * Roles to be used with the z_bind()
 */
#define Z_ORIGIN         1
#define Z_TARGET         2


/*
 * Origin/Target states for the progress of the Z-association
 */
#define Z_UNINIT             0 /* not yet uninitialized                      */
#define Z_UNBOUND            1 /* not bound to a local address               */
#define Z_IDLE               2 /* not connected at lower layer               */
#define Z_INCON              3 /* Incoming connection pending                */
#define Z_OUTCON             4 /* Outgoing connection pending                */
#define Z_CLOSED             5 /* origin is awaiting an Init request from    */
			       /* user or target is awaiting an Init PDU     */
#define Z_INIT_SENT          6 /* origin is awaiting an Init response PDU    */
#define Z_INIT_RECVD         7 /* target is awaiting an Init response from   */
                               /* user                                       */
#define Z_ACC_SENT           8 /* During initialization target has sent an   */
                               /* Access-control PDU and is awaiting an      */
                               /* Access-control-response PDU from origin    */
#define Z_ACC_RECVD          9 /* During initialization origin has received  */
                               /* an Access-control PDU and is awaiting an   */
                               /* Access-control-response from user          */
#define Z_RSC_SENT          10 /* During initialization target has sent      */
                               /* a Resource-control PDU and is awaiting a   */
                               /* Resource-control-response PDU from origin  */
#define Z_RSC_RECVD         11 /* During initialization origin has received  */
                               /* a Resource-control PDU and is awaiting a   */
                               /* Resource-control-response from user        */
#define Z_SERIAL_IDLE       12 /* Z-association is established, there are no */
                               /* active operations and 'serial operations'  */
                               /* is in effect                               */
#define Z_CONCURRENT_IDLE   13 /* Z-association is established, there are no */
                               /* active operations and 'concurrent          */
                               /* operations' is in effect                   */
#define Z_SERIAL_ACTIVE     14 /* There is an active operation and 'serial   */
                               /* operations' is in effect                   */
#define Z_CONCURRENT_ACTIVE 15 /* There is at least one active operation and */
                               /* 'concurrent operations' is in effect       */
#define Z_Z_ACC_SENT        16 /* Target has sent an Access-control PDU and  */
                               /* is awaiting an Access-control-response PDU */
                               /* from origin                                */
#define Z_Z_ACC_RECVD       17 /* Origin has received an Access-control PDU  */
                               /* and is awaiting an Access-control-response */
                               /* from user                                  */
#define Z_Z_RSC_SENT        18 /* Target has sent a Resource-control PDU and */
                               /* is awaiting a resource-control-response    */
                               /* PDU from origin                            */
#define Z_Z_RSC_RECVD       19 /* Origin has received a Resource-control PDU */
                               /* and is awaiting a Resource-control-response*/
                               /* from user                                  */
#define Z_CLOSE_SENT        20 /* Origin/Target is awaiting a Close PDU from */
                               /* Target/Origin                              */
#define Z_CLOSE_RECVD       21 /* Origin/Target is awaiting a Close response */
                               /* from user                                  */
#define Z_PRESENT_SENT      22 /* Origin is awaiting a Present response PDU  */
                               /* from Target                                */
#define Z_PRESENT_RECVD     23 /* Target is awaiting a Present response from */
                               /* user                                       */


/*
 * Definitions for all the Z39.50 primitives
 * used by both the z_snd() && z_rcv() routines.
 */


/*
 * Z39.50 operations: defined in terms of Action(s) and Event(s).
 * Origin/Target Events for Z-association
 */
#define Z_BADEVENT       (-1)
#define Z_NOEVENT      (0x00)
/*
 * Lower Layer events to handle the establishment of the Z-associations
 */
#define Z_LISTEN       (0x01)      /* connection indication received */
#define Z_CONNECT      (0x02)      /* connect confirmation received  */
#define Z_DISCON       (0x03)      /* disconnect received            */
#define Z_GODATA       (0x04)      /* sending data is again possible */
/*
 * Init Request
 */
#define Z_INIT_REQ     (0x05)
#define Z_INIT_IND     (0x06)
/*
 * Init Response
 */
#define Z_INIT_RSP     (0x07)
#define Z_INIT_CNF     (0x08)
/*
 * Search Request
 */
#define Z_SEARCH_REQ   (0x09)
#define Z_SEARCH_IND   (0x0a)
/*
 * Search Response
 */
#define Z_SEARCH_RSP   (0x0b)
#define Z_SEARCH_CNF   (0x0c)
/*
 * Present Request
 */
#define Z_PRESENT_REQ  (0x0d)
#define Z_PRESENT_IND  (0x0e)
/*
 * Present Response
 */
#define Z_PRESENT_RSP  (0x0f)
#define Z_PRESENT_CNF  (0x10)
/*
 * Delete Request
 */
#define Z_DELETE_REQ   (0x11)
#define Z_DELETE_IND   (0x12)
/*
 * Delete Response
 */
#define Z_DELETE_RSP   (0x13)
#define Z_DELETE_CNF   (0x14)
/*
 * Access Control Request
 */
#define Z_ACCESS_REQ   (0x15)
#define Z_ACCESS_IND   (0x16)
/*
 * Access Control Response
 */
#define Z_ACCESS_RSP   (0x17)
#define Z_ACCESS_CNF   (0x18)
/*
 * Resource Control Request
 */
#define Z_RESOURCE_REQ (0x19)
#define Z_RESOURCE_IND (0x1a)
/*
 * Resource Control Response
 */
#define Z_RESOURCE_RSP (0x1b)
#define Z_RESOURCE_CNF (0x1c)
/*
 * Trigger Resource Control Request
 */
#define Z_TRIGGER_REQ  (0x1d)
#define Z_TRIGGER_IND  (0x1e)
/*
 * Resource Report Request
 */
#define Z_REPORT_REQ   (0x1f)
#define Z_REPORT_IND   (0x20)
/*
 * Resource Report Response
 */
#define Z_REPORT_RSP   (0x21)
#define Z_REPORT_CNF   (0x22)
/*
 * Scan Request
 */
#define Z_SCAN_REQ     (0x23)
#define Z_SCAN_IND     (0x24)
/*
 * Scan Response
 */
#define Z_SCAN_RSP     (0x25)
#define Z_SCAN_CNF     (0x26)
/*
 * Sort Request
 */
#define Z_SORT_REQ     (0x27)
#define Z_SORT_IND     (0x28)
/*
 * Sort Response
 */
#define Z_SORT_RSP     (0x29)
#define Z_SORT_CNF     (0x2a)
/*
 * Segment Request
 */
#define Z_SEGMENT_REQ  (0x2b)
#define Z_SEGMENT_IND  (0x2c)
/*
 * Extended Services Request
 */
#define Z_EXTENDED_REQ (0x2d)
#define Z_EXTENDED_IND (0x2e)
/*
 * Extended Services Response
 */
#define Z_EXTENDED_RSP (0x2f)
#define Z_EXTENDED_CNF (0x30)
/*
 * Close
 */
#define Z_CLOSE_REQ    (0x31)
#define Z_CLOSE_IND    (0x32)
#define Z_CLOSE_RSP    Z_CLOSE_REQ
#define Z_CLOSE_CNF    Z_CLOSE_IND



/*
 * -=-=-=-=-=-=-=-=-=-
 * -= General types =-
 * -=-=-=-=-=-=-=-=-=-
 */


/*
 * Octet String.
 * Definition used to carry service parameters with no semantics
 * (trasparent binary data).
 *
 * Optional parameters can be specified by setting the size to -1.
 * It indicates that the parameter is not present.
 */
typedef struct
{
  int    size;
  char * data;
} zoctetstring;


/*
 * ANSI has assigned the following Object Identifier to Z39.50 standard:
 *
 * {iso member-body US ANSI-standard-Z39.50}
 * that is, using the shorthand notation,
 * {1 2 840 10003}
 *
 * Appendix 1 of ANSI/NISO Z39.50-1995 defines the complete set of
 * Object Identifiers at the level subordinate to the previous
 * definition.
 *
 * The definition for the Object Identifiers is given with respect
 * to the user. In this implementation of the Z39.50 standard each
 * Object Identifier simple consists in a string in dot notation.
 *
 * So all Z39.50 Object Identifiers must be prefixed with the string
 * "1.2.840.10003" that summarizes valid Z39.50 OIDs in dot notation.
 *
 */

typedef char zobjectid;


/*
 * -=-=-=-=-=-=-=-=-=-=-=
 * -= Isomorphic types -=
 * -=-=-=-=-=-=-=-=-=-=-=
 */
typedef char                 zvisiblestring;
typedef char                 zgraphicstring;
typedef char                 zinternationalstring;

typedef zinternationalstring zelementsetname;
typedef zinternationalstring zdatabasename;

typedef zoctetstring         zbitstring;
typedef zvisiblestring       zgeneralizedtime;
typedef zoctetstring         zanytype;
typedef zoctetstring         zreferenceid;
typedef zobjectid            zattributesetid;
typedef zinternationalstring zresultsetid;
typedef zobjectid            zresourcereportid;


/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -= Transfer Syntaxes Identifiers =-
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 */
#define Z_BER                  "2.1.1"
#define Z_ISO2709              "1.0.2709.1.1"

/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -= Attribute Set Identifiers =-
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 * Appendix 3 of ANSI/NISO Z39.50-1995 registers the following:
 */
#define Z_BIB_1                "1.2.840.10003.3.1"
#define Z_EXP_1                "1.2.840.10003.3.2"
#define Z_EXT_1                "1.2.840.10003.3.3"
#define Z_CCL_1                "1.2.840.10003.3.4"
#define Z_GILS                 "1.2.840.10003.3.5"
#define Z_STAS                 "1.2.840.10003.3.6"
#define Z_CIMI_1               "1.2.840.10003.3.8"

/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -= Diagnostic Set and Format Identifiers =-
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 * Appendix 4 of ANSI/NISO Z39.50-1995 registers the following
 * object identifiers for use with the diagnostic records:
 */
#define Z_DIAGSET_BIB_1        "1.2.840.10003.4.1"
#define Z_DIAGFORMAT_DIAG_1    "1.2.840.10003.4.2"

/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -= Record Syntax Identifiers =-
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 * Appendix 5 of ANSI/NISO Z39.50-1995 defines the following
 * object identifiers for use with the record syntaxes:
 */
#define Z_UNIMARC              "1.2.840.10003.5.1"
#define Z_INTERMARC            "1.2.840.10003.5.2"
#define Z_CCF                  "1.2.840.10003.5.3"
#define Z_USMARC               "1.2.840.10003.5.10"
#define Z_UKMARC               "1.2.840.10003.5.11"
#define Z_NORMARC              "1.2.840.10003.5.12"
#define Z_LIBRISMARC           "1.2.840.10003.5.13"
#define Z_DANMARC              "1.2.840.10003.5.14"
#define Z_FINMARC              "1.2.840.10003.5.15"
#define Z_MAB                  "1.2.840.10003.5.16"
#define Z_CANMARC              "1.2.840.10003.5.17"
#define Z_SBN                  "1.2.840.10003.5.18"
#define Z_PICAMARC             "1.2.840.10003.5.19"
#define Z_AUSMARC              "1.2.840.10003.5.20"
#define Z_IBERMARC             "1.2.840.10003.5.21"

#define Z_EXPLAIN              "1.2.840.10003.5.100"
#define Z_SUTRS                "1.2.840.10003.5.101"
#define Z_OPAC                 "1.2.840.10003.5.102"
#define Z_SUMMARY              "1.2.840.10003.5.103"

#define Z_GRS_1                "1.2.840.10003.5.105"
#define Z_EXTENDED_SERVICES    "1.2.840.10003.5.106"


/*
 * Object Identifiers in the first range [1-21] (so called bibliographic
 * syntaxes) are not described via ASN.1 and they may be used in
 * conjuncion with the Transfer Syntax ISO2709 "1.2.2709.1.1"
 * {iso standard 2709 transfer-syntax character-encoding} for tranfer
 * of bibliographic records.
 *
 * Other Object Identifiers in the range [100-106] are described via ASN.1.
 * ISO8825 BER Transfer Syntax "1.2.1" {iso joint-iso-ccitt basic-encoding}
 * may be used for tranfer of such records.
 */

/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 * -= Resource-Report-Format Identifiers -=
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 */
#define Z_RESOURCE1            "1.2.840.10003.7.1"
#define Z_RESOURCE2            "1.2.840.10003.7.2"


/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 * -= Access-Report-Format Identifiers -=
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 */
#define Z_PROMPT1              "1.2.840.10003.8.1"
#define Z_DES1                 "1.2.840.10003.8.2"
#define Z_KRB1                 "1.2.840.10003.8.3"


/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * -= Extended-Services Identifiers =-
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 */
#define Z_PERSISTENT_RESULTSET "1.2.840.10003.9.1"
#define Z_PERSISTENT_QUERY     "1.2.840.10003.9.2"
#define Z_PERIODIC_QUERY       "1.2.840.10003.9.3"
#define Z_ITEM_ORDER           "1.2.840.10003.9.4"
#define Z_DATABASE_UPDATE      "1.2.840.10003.9.5"
#define Z_EXPORT_SPECIFICATION "1.2.840.10003.9.6"
#define Z_EXPORT_INVOCATION    "1.2.840.10003.9.7"


#define Z_DIGITAL_COLL_SCHEMA  "1.2.840.10003.13.3"
#define Z_CIMI_SCHEMA          "1.2.840.10003.13.5"
#define Z_CIMI_TYPE_OF_DESCR   "2"
#define Z_CIMI_TYPE_OF_OBJ     "1"
#define Z_CIMI_CATEG_OF_OBJ    "cimi: object record"

#define Z_TAGSET_M             "1.2.840.10003.14.1"
#define Z_TAGSET_G             "1.2.840.10003.14.2"
#define Z_TAGSET_DIGITAL_COLL  "1.2.840.10003.14.5"
#define Z_TAGSET_CIMI          "1.2.840.10003.14.6"


#define EMPTY_ELEMENT          "emptyElement"

/*
 * Single ASN.1 external types (both generic and specific to Z39.50 Protocol).
 */


/*
 * Generic external types.
 */
#define Z_SINGLE_ASN1_TYPE_ENCODING  1
#define Z_OCTET_ALIGNED_ENCODING     2
#define Z_ARBITRARY_ENCODING         3

/*
 * External structure.
 */

typedef struct
{
  int         size;
  zanytype ** table;    /* pointer to the table of pointers */
} zoctetaligned;


typedef struct
{
  zobjectid       * direct_reference;       /* OPTIONAL */
  int             * indirect_reference;     /* OPTIONAL */
  zgraphicstring  * data_value_descriptor;  /* OPTIONAL */
  int choice;
  union
  {
    zanytype      * single_asn1_type;
    zoctetaligned * octet_aligned;
    zbitstring    * arbitrary;
  } encoding;
} zexternal;


/*
 * Other information
 */

typedef struct
{
  zobjectid  * categorytypeid;         /* OPTIONAL */
  int          categoryvalue;
} zinfocategory;


#define Z_OTHERINFO_CHARACTERINFO         1
#define Z_OTHERINFO_BINARYINFO            2
#define Z_OTHERINFO_EXTERNALLYDEFINEDINFO 3
#define Z_OTHERINFO_OID                   4
typedef struct
{
  zinfocategory          * category;   /* OPTIONAL */
  int choice;
  union
  {
    zinternationalstring * characterinfo;
    zoctetstring         * binaryinfo;
    zexternal            * externallydefinedinfo;
    zobjectid            * oid;
  } u;
} zotherinformationelement;


typedef struct
{
  int                         size;
  zotherinformationelement ** table; /* pointer to the table of pointers */
} zotherinformation;





#define Z_STRING_STRINGORNUMERIC  1
#define Z_NUMBER_STRINGORNUMERIC  2
typedef struct
{
  int choice;
  union
  {
    zinternationalstring * string;
    int                    numeric;
  } u;
} zstringornumeric;




/*
 * -=-=-=-=-=-=-=-=-=-=
 * -=  INIT SERVICE  -=
 * -=-=-=-=-=-=-=-=-=-=
 */

/*
 * The Init service allows the origin to establish a Z-association
 */

/*
 * Z39.50 protocol version.
 *
 * Both the origin and the target indicate in the Initialize Request
 * and Response APDUs their supported versions. The highest version in
 * common is 'in force' for the Z-association (if the Z-association
 * has been established), or none is in effect (and the Z-association)
 * is rejected.
 *
 * Note:
 * o V1 should be always set but it does not correspond to any Z39.50 version.
 * o V1 and V2 are identical. This implementation will also set V1 if the
 *   user has forgot reccomendation for interoperability with
 *   implementations ISO 10163-1991 (supporting version 1 only).
 */
#define Z_PROTOCOL_V1  (0x0001)
#define Z_PROTOCOL_V2  (0x0002)
#define Z_PROTOCOL_V3  (0x0004)
#define Z_PROTOCOL_ALL (Z_PROTOCOL_V1|Z_PROTOCOL_V2|Z_PROTOCOL_V3)

/*
 * Z39.50 protocol options.
 *
 * The origin proposes in the Init Service the set of capabilities desidered
 * and the response of the target determines which of them are 'in force'
 * for the Z-association.
 *
 * To increase interoperability any of the following options:
 * search, present, delete, resource report, scan, sort, extended services
 * may be negotiated for any version (even ANSI Z39.50-1992 version 2
 * did not define them).
 *
 * Warning:
 *   Level-1 Segmentation, Level-2 Segmentation and Concurrent Operations
 *   are available only when version 3 is proposed (or negotiated)
 */
#define ZO_SEARCH                 (0x000001)
#define ZO_PRESENT                (0x000002)
#define ZO_DELETE_RESULT_SET      (0x000004)
#define ZO_RESOURCE_REPORT        (0x000008)
#define ZO_TRIGGER_RESOURCE_CTRL  (0x000010)
#define ZO_RESOURCE_CTRL          (0x000020)
#define ZO_ACCESS_CTRL            (0x000040)
#define ZO_SCAN                   (0x000080)
#define ZO_SORT                   (0x000100)
#define ZO_RESERVED               (0x000200)
#define ZO_EXTENDED_SERVICES      (0x000400)
#define ZO_LEVEL_1_SEGMENTATION   (0x000800)
#define ZO_LEVEL_2_SEGMENTATION   (0x001000)
#define ZO_CONCURRENT_OPERATIONS  (0x002000)
#define ZO_NAMED_RESULT_SETS      (0x004000)
#define ZO_OPTION_ALL             (0x007fff)

/*
 * The subset of the protocol options that
 * can be proposed with the Z_INIT_REQ and Z_INIT_RSP
 * (and the subset of those currently unsupported)
 */
#define Z_ALLOWED_OPTIONS \
 (ZO_SEARCH               | ZO_PRESENT               | \
  ZO_DELETE_RESULT_SET    | ZO_SCAN                  | \
  ZO_SORT                 | ZO_NAMED_RESULT_SETS)

#define Z_UNSUPPORTED_OPTIONS \
 (ZO_RESOURCE_REPORT      | ZO_TRIGGER_RESOURCE_CTRL | \
  ZO_RESOURCE_CTRL        | ZO_ACCESS_CTRL           | \
  ZO_EXTENDED_SERVICES    | ZO_LEVEL_1_SEGMENTATION  | \
  ZO_LEVEL_2_SEGMENTATION | ZO_CONCURRENT_OPERATIONS)


/*
 * -=-=-=-=-=-=-=-=-=-=-=
 * -=  AUTHENTICATION  -=
 * -=-=-=-=-=-=-=-=-=-=-=
 */

/*
 * The definition of  `zidauthentication' is given to improve portability.
 *
 * If the origin and target agree on the use of the authentication mechanism
 * a value must be supplied by the origin in the Init Service and it is
 * used by the target to determine if the communication can be established.
 *
 * It corresponds to the following ASN.1 definition:
 *   IdAuthentication [7] CHOICE {
 *   open      VisibleString,
 *   idPass    SEQUENCE {
 *   groupId   [0]  IMPLICIT InternationalString OPTIONAL,
 *   userId    [1]  IMPLICIT InternationalString OPTIONAL,
 *   password  [2]  IMPLICIT InternationalString OPTIONAL },
 *   anonymous NULL,
 *   other     EXTERNAL }
 *
 * For compatibility with earlier versions the authentication parameter
 * is carried in the protocol by mean of `zanytype' object.
 */


/*
 * AUTHENTICATION TYPE.
 */
#define Z_AUTHTYPE_OPEN        1
#define Z_AUTHTYPE_IDPASSWORD  2
#define Z_AUTHTYPE_ANONYMOUS   3
#define Z_AUTHTYPE_OTHER       4

typedef struct
{
  zinternationalstring * groupid;    /* OPTIONAL */
  zinternationalstring * userid;     /* OPTIONAL */
  zinternationalstring * password;   /* OPTIONAL */
} zidpassword;


typedef struct
{
  int choice;
  union
  {
    zvisiblestring * open;
    zidpassword    * idpass;
    void           * anonymous;
    zexternal      * other;
  } u;
} zidauthentication;



typedef struct
{
  zreferenceid         * referenceid;             /* OPTIONAL */
  int                    protocolversion;         /* BITMASK */
  int                    options;                 /* BITMASK */
  unsigned long          preferredmessagesize;
  unsigned long          exceptionalrecordsize;
  zidauthentication    * idauthentication;        /* OPTIONAL */
  zinternationalstring * implementationid;        /* OPTIONAL */
  zinternationalstring * implementationname;      /* OPTIONAL */
  zinternationalstring * implementationversion;   /* OPTIONAL */
  zexternal            * userinformationfield;    /* OPTIONAL */
  zotherinformation    * otherinfo;               /* OPTIONAL */
} zinitializerequest;


/*
 * Result of negotiation
 */
#define Z_ACCEPTED  (1)  /* Accepted by the target */
#define Z_REJECTED  (2)  /* Rejected by the target */

typedef struct
{
  zreferenceid         * referenceid;            /* OPTIONAL */
  int                    protocolversion;        /* BITMASK */
  int                    options;                /* BITMASK */
  unsigned long          preferredmessagesize;
  unsigned long          exceptionalrecordsize;
  int                    result;                 /* Z_ACCEPTED or Z_REJECTED */
  zinternationalstring * implementationid;       /* OPTIONAL */
  zinternationalstring * implementationname;     /* OPTIONAL */
  zinternationalstring * implementationversion;  /* OPTIONAL */
  zexternal            * userinformationfield;   /* OPTIONAL */
  zotherinformation    * otherinfo;              /* OPTIONAL */
} zinitializeresponse;


/*
 * -=-=-=-=-=-=-=-=-=-=-=
 * -=  SEARCH SERVICE  -=
 * -=-=-=-=-=-=-=-=-=-=-=
 */

/*
 * The Search service enables an origin to query databases at a target
 * system and to receive information about the result of the query.
 */

/*
 * Database.
 *
 * How a database definition list plays with memory !!!
 *
 *   database
 * +----------+
 * |   size   |
 * | -------- |   Table of the Database Elements
 * |  table   |-> +-------------+
 * +----------+   |             |-> Database Element[0]
 *                +-------------+
 *                |   ......    |
 *                +-------------+
 *                |             |-> Database Element[n-1]
 *                +-------------+
 *
 */
typedef struct
{
  int                     size;
  zinternationalstring ** table; /* pointer to the table of pointers */
} zdatabasenames;


/*
 * Database Specific Unit
 */
typedef struct
{
  zdatabasename   * dbname;
  zelementsetname * esn;
} zdatabasespecificunit;

typedef struct
{
  int                      size;
  zdatabasespecificunit ** table;
} zdatabasespecific;



#define Z_BRIEF  "B"
#define Z_FULL   "F"


#define Z_GENERIC                1
#define Z_DATABASESPECIFIC       2
typedef struct
{
  int choice;
  union
  {
    zinternationalstring * genericelementsetname;
    zdatabasespecific    * databasespecific;
  } u;
} zelementsetnames;


#define Z_UNITSYSTEM_UNIT   1
#define Z_UNITTYPE_UNIT     2
#define Z_UNIT_UNIT         3
#define Z_SCALEFACTOR_UNIT  4
typedef struct
{
  zinternationalstring * unitsystem;    /* OPTIONAL */
  zstringornumeric     * unittype;      /* OPTIONAL */
  zstringornumeric     * unit;          /* OPTIONAL */
  int                  * scalefactor;   /* OPTIONAL */
} zunit;


typedef struct
{
  int     value;
  zunit * unitused;
} zintunit;


/*
 * Z39.50 defines an attribute to carry extra information about a term
 * It indicates the sematic meaning to a term.
 */

/*
 * The set of attributes for each individual search term.
 */

/*
 * Individual search term definition.
 */
#define Z_GENERAL_TERM          1
#define Z_NUMERIC_TERM          2
#define Z_CHARACTERSTRING_TERM  3
#define Z_OID_TERM              4
#define Z_DATETIME_TERM         5
#define Z_EXTERNAL_TERM         6
#define Z_INTEGERANDUNIT_TERM   7
#define Z_NULL_TERM             8
typedef struct
{
  int choice;
  union
  {
    zoctetstring         * general;
    /*
     * values below may be used only if version 3 is in force
     */
    int                    numeric;
    zinternationalstring * characterstring;
    zobjectid            * oid;
    zgeneralizedtime     * datetime;
    zexternal            * external;
    zintunit             * integerandunit;
    char                   null;
  } u;
} zterm;


typedef struct
{
  int                 size;
  zstringornumeric ** table;
} zattrlist;


typedef struct
{
  int    size;
  int ** table;
} zsemanticaction;


typedef struct
{
  zattrlist       * list;
  zsemanticaction * action;     /* OPTIONAL */
} zattrcomplex;


#define Z_NUMERIC_ATTRVALUE  1
#define Z_COMPLEX_ATTRVALUE  2
typedef struct
{
  int choice;
  union
  {
    int            numeric;
    zattrcomplex * complex;
  } u;
} zattrvalue;


/*
 * Each attribute of an Attribute Set is used to indicate the characteristics
 * of a term when the query is specified in the form attributelist+term
 *
 * Each attribute element has a type and a value.
 */

/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * Start section related to the ATTRIBUTE SET BIB-1
 *
 * Definitions from:
 * ANSI/NISO Z39.50-1995 Appendix 3 ATR.1 Attribute Set bib-1
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 */

/*
 * BIB1 Attribute types from table ATR.1.1
 */
#define BIB1_USE           1
#define BIB1_RELATION      2
#define BIB1_POSITION      3
#define BIB1_STRUCTURE     4
#define BIB1_TRUNCATION    5
#define BIB1_COMPLETENESS  6
#define CIMI_AUTHORITY     101


/*
 * BIB1 Use Attributes from table ATR.A-3-1
 */
#define BIB1_PERSONAL_NAME                   1
#define BIB1_CORPORATE_NAME                  2
#define BIB1_CONFERENCE_NAME                 3
#define BIB1_TITLE                           4
#define BIB1_TITLE_SERIES                    5
#define BIB1_TITLE_UNIFORM                   6
#define BIB1_ISBN                            7
#define BIB1_ISSN                            8
#define BIB1_LC_CARD_NUMBER                  9
#define BIB1_BNB_CARD_NUMBER                10
#define BIB1_BGF_SIC_NUMBER                 11
#define BIB1_LOCAL_NUMBER                   12
#define BIB1_DEWEY_CLASSIFICATION           13
#define BIB1_UDC_CLASSIFICATION             14
#define BIB1_BLISS_CLASSIFICATION           15
#define BIB1_LC_CALL_NUMBER                 16
#define BIB1_NLM_CALL_NUMBER                17
#define BIB1_NAL_CALL_NUMBER                18
#define BIB1_MOS_CALL_NUMBER                19
#define BIB1_LOCAL_CLASSIFICATION           20
#define BIB1_SUBJECT_HEADING                21
#define BIB1_SUBJECT_RAMEAU                 22
#define BIB1_BDI_INDEX_SUBJECT              23
#define BIB1_INSPEC_SUBJECT                 24
#define BIB1_MESH_SUBJECT                   25
#define BIB1_PA_SUBJECT                     26
#define BIB1_LC_SUBJECT_HEADING             27
#define BIB1_RVM_SUBJECT_HEADING            28
#define BIB1_LOCAL_SUBJECT_INDEX            29
#define BIB1_DATE                           30
#define BIB1_DATE_OF_PUBLICATION            31
#define BIB1_DATE_OF_ACQUISITION            32
#define BIB1_TITLE_KEY                      33
#define BIB1_TITLE_COLLECTIVE               34
#define BIB1_TITLE_PARALLEL                 35
#define BIB1_TITLE_COVER                    36
#define BIB1_TITLE_ADDED_TITLE_PAGE         37
#define BIB1_TITLE_CAPTION                  38
#define BIB1_TITLE_RUNNING                  39
#define BIB1_TITLE_SPINE                    40
#define BIB1_TITLE_OTHER_VARIANT            41
#define BIB1_TITLE_FORMER                   42
#define BIB1_TITLE_ABBREVIATED              43
#define BIB1_TITLE_EXPANDED                 44
#define BIB1_SUBJECT_PRECIS                 45
#define BIB1_SUBJECT_RSWK                   46
#define BIB1_SUBJECT_SUBDIVISION            47
#define BIB1_NUMBER_NATL_BIBLIOGRAPHY       48
#define BIB1_NUMBER_LEGAL_DEPOSIT           49
#define BIB1_NUMBER_GOVT_PUBLICATION        50
#define BIB1_NUMBER_PUBLISHER_FOR_MUSIC     51
#define BIB1_NUMBER_DB                      52
#define BIB1_NUMBER_LOCAL_CALL              53
#define BIB1_CODE_LANGUAGE                  54
#define BIB1_CODE_GEOGRAPHIC_AREA           55
#define BIB1_CODE_INSTITUTION               56
#define BIB1_NAME_AND_TITLE                 57
#define BIB1_NAME_GEOGRAPHIC                58
#define BIB1_PLACE_PUBLICATION              59
#define BIB1_CODEN                          60
#define BIB1_MICROFORM_GENERATION           61
#define BIB1_ABSTRACT                       62
#define BIB1_NOTE                           63
#define BIB1_AUTHOR_TITLE                 1000
#define BIB1_RECORD_TYPE                  1001
#define BIB1_NAME                         1002
#define BIB1_AUTHOR                       1003
#define BIB1_AUTHOR_NAME_PERSONAL         1004
#define BIB1_AUTHOR_NAME_CORPORATE        1005
#define BIB1_AUTHOR_NAME_CONFERENCE       1006
#define BIB1_IDENTIFIER_STANDARD          1007
#define BIB1_SUBJECT_LC_CHILDRENS         1008
#define BIB1_SUBJECT_NAME_PERSONAL        1009
#define BIB1_BODY_OF_TEXT                 1010
#define BIB1_DATE_TIME_ADDED_TO_DATABASE  1011
#define BIB1_DATE_TIME_LAST_MODIFIED      1012
#define BIB1_AUTHORITY_FORMAT_IDENTIFIER  1013
#define BIB1_CONCEPT_TEXT                 1014
#define BIB1_CONCEPT_REFERENCE            1015
#define BIB1_ANY                          1016
#define BIB1_SERVER_CHOICE                1017
#define BIB1_PUBLISHER                    1018
#define BIB1_RECORD_SOURCE                1019
#define BIB1_EDITOR                       1020
#define BIB1_BIB_LEVEL                    1021
#define BIB1_GEOGRAPHIC_CLASS             1022
#define BIB1_INDEXED_BY                   1023
#define BIB1_MAP_SCALE                    1024
#define BIB1_MUSIC_KEY                    1025
#define BIB1_RELATED_PERIODICAL           1026
#define BIB1_REPORT_NUMBER                1027
#define BIB1_STOCK_NUMBER                 1028
#define BIB1_THEMATIC_NUMBER              1030
#define BIB1_MATERIAL_TYPE                1031
#define BIB1_DOC_ID                       1032
#define BIB1_HOST_ITEM                    1033
#define BIB1_CONTENT_TYPE                 1034
#define BIB1_ANYWHERE                     1035
#define BIB1_AUTHOR_TITLE_SUBJECT         1036

#define CIMI1_IMAGE                       2020

/*
 * BIB1 Relation Attributes from table ATR.A-3-2
 */
#define BIB1_RELATION_LESSTHAN               1
#define BIB1_RELATION_LESSTHANOREQUAL        2
#define BIB1_RELATION_EQUAL                  3
#define BIB1_RELATION_GREATEROREQUAL         4
#define BIB1_RELATION_GREATERTHAN            5
#define BIB1_RELATION_NOTEQUAL               6
#define BIB1_RELATION_PHONETIC             100
#define BIB1_RELATION_STEM                 101
#define BIB1_RELATION_RELEVANCE            102
#define BIB1_RELATION_ALWAYSMATCHES        103


/*
 * BIB1 Position Attributes from table ATR.A-3-3
 */
#define BIB1_POSITION_FIRST_IN_FIELD         1
#define BIB1_POSITION_FIRST_IN_SUBFIELD      2
#define BIB1_POSITION_ANY_POSITION_IN_FIELD  3


/*
 * BIB1 Structure Attributes from table ATR.A-3-4
 */
#define BIB1_STRUCTURE_PHRASE              1
#define BIB1_STRUCTURE_WORD                2
#define BIB1_STRUCTURE_KEY                 3
#define BIB1_STRUCTURE_YEAR                4
#define BIB1_STRUCTURE_DATE_1              5
#define BIB1_STRUCTURE_WORD_LIST           6
#define BIB1_STRUCTURE_DATE_2            100
#define BIB1_STRUCTURE_NAME_1            101
#define BIB1_STRUCTURE_NAME_2            102
#define BIB1_STRUCTURE_STRUCTURE         103
#define BIB1_STRUCTURE_URX               104
#define BIB1_STRUCTURE_FREE_FORM_TEXT    105
#define BIB1_STRUCTURE_DOCUMENT_TEXT     106
#define BIB1_STRUCTURE_LOCAL_NUMBER      107
#define BIB1_STRUCTURE_STRING            108
#define BIB1_STRUCTURE_NUMERIC_STRING    109

/*
 * BIB1 Truncation Attributes from table ATR.A-3-5
 */
#define BIB1_TRUNCATION_RIGHT                1
#define BIB1_TRUNCATION_LEFT                 2
#define BIB1_TRUNCATION_RIGHT_AND_LEFT       3
#define BIB1_TRUNCATION_DO_NOT_TRUNCATE    100
#define BIB1_TRUNCATION_PROCESS_IN_SEARCH  101
#define BIB1_TRUNCATION_REG_EXPR_1         102
#define BIB1_TRUNCATION_REG_EXPR_2         103


/*
 * BIB1 Completeness Attributes from table ATR.A-3-6
 */
#define BIB1_COMPLETENESS_INCOMPLETE_SUBFIELD  1
#define BIB1_COMPLETENESS_COMPLETE_SUBFIELD    2
#define BIB1_COMPLETENESS_COMPLETE_FIELD       3


/*
 * Diagnostic set bib-1.
 * Defined in ANSI/NISO Z39.50-1995 Appendix 4 - Table A-4-1
 * on page 85 and following.
 */
#define BIB1_DIAG_1      1 /* permanent system error */
#define BIB1_DIAG_2      2 /* temporary system error */
#define BIB1_DIAG_3      3 /* unsupported search */
#define BIB1_DIAG_4      4 /* Terms only exclusion (stop) words */
#define BIB1_DIAG_5      5 /* Too many argument words */
#define BIB1_DIAG_6      6 /* Too many boolean operators */
#define BIB1_DIAG_7      7 /* Too many truncated words */
#define BIB1_DIAG_8      8 /* Too many incomplete subfields */
#define BIB1_DIAG_9      9 /* Truncated words too short */
#define BIB1_DIAG_10    10 /* Invalid format for rec. number (search term) */
#define BIB1_DIAG_11    11 /* Too many characters in search statement */
#define BIB1_DIAG_12    12 /* Too many records retrieved */
#define BIB1_DIAG_13    13 /* Present request out-of-range */
#define BIB1_DIAG_14    14 /* System error in presenting records */
#define BIB1_DIAG_15    15 /* Record not authorized to be sent intersystem */
#define BIB1_DIAG_16    16 /* Record exceeds Preferred-message-size */
#define BIB1_DIAG_17    17 /* Record exceeds Maximum-record-size */
#define BIB1_DIAG_18    18 /* Result set not supported as a search term */
#define BIB1_DIAG_19    19 /* Only single result set as search term supported */
#define BIB1_DIAG_20    20
#define BIB1_DIAG_21    21 /* Result set exists and replace indicator off */
#define BIB1_DIAG_22    22 /* Result set naming not supported */
#define BIB1_DIAG_23    23 /* Specified combination of databases not supported */
#define BIB1_DIAG_24    24 /* Element set names not supported */
#define BIB1_DIAG_25    25
#define BIB1_DIAG_26    26 /* Only generic form of element set name supported  */
#define BIB1_DIAG_27    27 /* Result set no longer exists - unilaterally deleted by target  */
#define BIB1_DIAG_28    28 /* Result set is in use  */
#define BIB1_DIAG_29    29 /* One of the specified databases is locked */
#define BIB1_DIAG_30    30 /* Specified result set does not exist */
#define BIB1_DIAG_31    31 /* Resources exhausted - no results available */
#define BIB1_DIAG_32    32 /* Resources exhausted - unpred. partial results available  */
#define BIB1_DIAG_33    33 /* Resources exhausted - valid subset of results available */

#define BIB1_DIAG_100  100 /* (unspecified) error */
#define BIB1_DIAG_101  101 /* Access-control failure */
#define BIB1_DIAG_102  102 /* Challenge required, could not be issued - op. terminated  */
#define BIB1_DIAG_103  103 /* Challenge required, could not be issued - rec. not included  */
#define BIB1_DIAG_104  104 /* Challenge failed - record not included */
#define BIB1_DIAG_105  105 /* Terminated at origin request */
#define BIB1_DIAG_106  106
#define BIB1_DIAG_107  107 /* Query type not supported */
#define BIB1_DIAG_108  108 /* Malformed query */
#define BIB1_DIAG_109  109 /* Database unavailable */
#define BIB1_DIAG_110  110 /* Operator unsupported */
#define BIB1_DIAG_111  111 /* Too many databases specified */
#define BIB1_DIAG_112  112 /* Too many result sets created */
#define BIB1_DIAG_113  113 /* Unsupported attribute type */
#define BIB1_DIAG_114  114 /* Unsupported Use attribute */
#define BIB1_DIAG_115  115 /* Unsupported value for Use attribute */
#define BIB1_DIAG_116  116 /* Use attribute required but not supplied */
#define BIB1_DIAG_117  117 /* Unsupported Relation attribute */
#define BIB1_DIAG_118  118 /* Unsupported Structure attribute */
#define BIB1_DIAG_119  119 /* Unsupported Position attribute */
#define BIB1_DIAG_120  120 /* Unsupported Truncation attribute */
#define BIB1_DIAG_121  121 /* Unsupported Attribute Set */
#define BIB1_DIAG_122  122 /* Unsupported Completeness attribute */
#define BIB1_DIAG_123  123 /* Unsupported attribute combination */
#define BIB1_DIAG_124  124 /* Unsupported coded value for term */
#define BIB1_DIAG_125  125 /* Malformed search term */
#define BIB1_DIAG_126  126 /* Illegal term value for attribute */
#define BIB1_DIAG_127  127 /* Unparsable format for un-normalized value */
#define BIB1_DIAG_128  128 /* Illegal result set name */
#define BIB1_DIAG_129  129 /* Proximity search of sets not supported */
#define BIB1_DIAG_130  130 /* Illegal result set in proximity search */
#define BIB1_DIAG_131  131 /* Unsupported proximity relation */
#define BIB1_DIAG_132  132 /* Unsupported proximity unit code */

#define BIB1_DIAG_201  201 /* Proximity not supported with this attribute combination  */
#define BIB1_DIAG_202  202 /* Unsupported distance for proximity */
#define BIB1_DIAG_203  203 /* Ordered flag not supported for proximity */

#define BIB1_DIAG_205  205 /* Only zero step size supported for Scan */
#define BIB1_DIAG_206  206
#define BIB1_DIAG_207  207 /* Cannot sort according to sequence */
#define BIB1_DIAG_208  208 /* No result set name supplied on Sort */
#define BIB1_DIAG_209  209 /* Generic sort not supported (db specific only)  */
#define BIB1_DIAG_210  210 /* Db specific sort not supported */
#define BIB1_DIAG_211  211 /* Too many sort keys */
#define BIB1_DIAG_212  212 /* Duplicate sort keys */
#define BIB1_DIAG_213  213 /* Unuspported missing data action */
#define BIB1_DIAG_214  214 /* Illegal sort relation */
#define BIB1_DIAG_215  215 /* Illegal case value */
#define BIB1_DIAG_216  216 /* Illegal missing data action */
#define BIB1_DIAG_217  217 /* Cannot guarantee records will fit in specified segments */
#define BIB1_DIAG_218  218 /* ES: Package name already in use */
#define BIB1_DIAG_219  219 /* ES: no such package, on modify/delete */
#define BIB1_DIAG_220  220 /* ES: quota exceeded */
#define BIB1_DIAG_221  221 /* ES: extended service type not supported */
#define BIB1_DIAG_222  222 /* ES: permission denied on ES - id not authorized */
#define BIB1_DIAG_223  223 /* ES: permission denied on ES - cannot modify or delete */
#define BIB1_DIAG_224  224 /* ES: immediate execution failed */
#define BIB1_DIAG_225  225 /* ES: immediate execution not supported for this service */
#define BIB1_DIAG_226  226 /* ES: immediate execution not supported for these parameters */
#define BIB1_DIAG_227  227 /* No data available in requested record syntax */
#define BIB1_DIAG_228  228 /* Scan: malformed scan */
#define BIB1_DIAG_229  229 /* Term type not supported */
#define BIB1_DIAG_230  230 /* Sort: too many input results */
#define BIB1_DIAG_231  231 /* Sort: incompatible record formats */
#define BIB1_DIAG_232  232
#define BIB1_DIAG_233  233
#define BIB1_DIAG_234  234
#define BIB1_DIAG_235  235
#define BIB1_DIAG_236  236
#define BIB1_DIAG_237  237
#define BIB1_DIAG_238  238
#define BIB1_DIAG_239  239
#define BIB1_DIAG_240  240
#define BIB1_DIAG_241  241
#define BIB1_DIAG_242  242
#define BIB1_DIAG_243  243
#define BIB1_DIAG_244  244
#define BIB1_DIAG_245  245 /* RPN1: restriction ('resultAttr') operand not supported */
#define BIB1_DIAG_246  246 /* RPN1: 'complex' attributeValue not supported */
#define BIB1_DIAG_247  247 /* RPN1: 'attributeSet' as part of AttributeElement unsupported */

#define BIB1_DIAG_1024 1024


/*
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 * End section related to the ATTRIBUTE SET BIB-1
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 */




typedef struct
{
  zattributesetid * attributeset;    /* OPTIONAL */
                                     /* Must be omitted if version 2 is in  */
                                     /* force. If included, overrides value */
                                     /* of attributeset                     */
  int               attributetype;
  zattrvalue      * attributevalue;
} zattributeelement;


/*
 * The table of attribute elements
 */
typedef struct
{
  int                   size;
  zattributeelement ** table;    /* pointer to the table of pointers */
} zattributelist;


typedef struct
{
  zattributelist * attributes;   /* List of attributes */
  zterm          * term;         /* Term */
} zattributesplusterm;


typedef struct
{
  zresultsetid   * resultset;
  zattributelist * attrlist;
} zresultsetplusattributes;


#define Z_LESSTHAN_RELATIONTYPE            1
#define Z_LESSTHANOREQUAL_RELATIONTYPE     2
#define Z_EQUAL_RELATIONTYPE               3
#define Z_GREATERTHANOREQUAL_RELATIONTYPE  4
#define Z_GREATERTHAN_RELATIONTYPE         5
#define Z_NOTEQUAL_RELATIONTYPE            6


#define Z_KNOWN_PROXIMITYUNITCODE          1
#define Z_PRIVATE_PROXIMITYUNITCODE        2

#define Z_KNOWNPROXIMITYUNIT_CHARACTER     1
#define Z_KNOWNPROXIMITYUNIT_WORD          2
#define Z_KNOWNPROXIMITYUNIT_SENTENCE      3
#define Z_KNOWNPROXIMITYUNIT_PARAGRAPH     4
#define Z_KNOWNPROXIMITYUNIT_SECTION       5
#define Z_KNOWNPROXIMITYUNIT_CHAPTER       6
#define Z_KNOWNPROXIMITYUNIT_DOCUMENT      7
#define Z_KNOWNPROXIMITYUNIT_ELEMENT       8
#define Z_KNOWNPROXIMITYUNIT_SUBELEMENT    9
#define Z_KNOWNPROXIMITYUNIT_ELEMENTTYPE  10
#define Z_KNOWNPROXIMITYUNIT_BYTE         11

typedef struct
{
  int  * exclusion;      /* OPTIONAL: Z_FALSE or Z_TRUE */
  int    distance;
  int    ordered;        /* Z_FALSE or Z_TRUE */
  int    relationtype;
  struct
  {
    int choice;
    union
    {
      int known;
      int private;
    } u;
  } zproximityunitcode;
} zproximityoperator;



/*
 * Operators.
 */
#define Z_AND_OPERATOR        1
#define Z_OR_OPERATOR         2
#define Z_ANDNOT_OPERATOR     3
#define Z_PROXIMITY_OPERATOR  4
typedef struct
{
  int choice;
  union
  {
    int                and;
    int                or;
    int                and_not;
    /*
     * If version 2 is in force:
     * prox may be used only if query type is 101
     */
    zproximityoperator * prox;
  } u;
} zoperator;



/*
 * Operand
 */
#define Z_ATTR_PLUS_TERM_OPERAND       1
#define Z_RESULTSET_OPERAND            2
#define Z_RESULTSET_PLUS_ATTR_OPERAND  3
typedef struct
{
  int choice;
  union
  {
    zattributesplusterm      * attrterm;
    zresultsetid             * resultset;
    zresultsetplusattributes * resultattr;
  } u;
} zoperand;


typedef struct
{
  struct zrpnstructure * rpn1;
  struct zrpnstructure * rpn2;
  zoperator            * op;
} zrpnrpnop;


#define Z_OP_RPNSTRUCTURE        1
#define Z_RPNRPNOP_RPNSTRUCTURE  2
typedef struct zrpnstructure
{
  int choice;
  union
  {
    zoperand  * single;   /* a single access point clause                   */
    zrpnrpnop * several;  /* or several clauses linked by logical operators */
  } u;
} zrpnstructure;



/*
 * Individual search terms are linked by boolean operators.
 * Terms and operators are expressed in Reverse Polish Notation.
 * Attributes associated with a search term belong to a registered
 * attribute set identifier (an Object Identifier) included with the query.
 *
 * For example the attribute set BIB-1 defines various attributes useful
 * for bibliographic queries. Legal values for the Attribute Set Id are
 * the object identifiers registered in Appendix 3 of ANSI/NISO Z39.50-1995.
 */
typedef struct
{
  zattributesetid * attributeset;
  zrpnstructure   * rpn;
} zrpnquery;



/*
 * Query formulation.
 * Terms may be combined in different types of queries.
 */
#define Z_QUERYTYPE_0     1   /* a priori agreement outside the standard */
#define Z_QUERYTYPE_1     2   /* RPN (Reverse Polish Notation) */
#define Z_QUERYTYPE_2     3   /* ISO 8777 */
#define Z_QUERYTYPE_100   4   /* ANSI Z39.58 */
#define Z_QUERYTYPE_101   5   /* ERPN (Extended Reverse Polish Notation) */
#define Z_QUERYTYPE_102   6   /* Ranked List */


/*
 * WARNING: Only query type 1 and 101 are currently supported.
 */
typedef struct
{
  int choice;
  union
  {
    zanytype     * type_0;       /* NOT SUPPORTED YET */
    zrpnquery    * type_1;
    zoctetstring * type_2;       /* NOT SUPPORTED YET */
    zoctetstring * type_100;     /* NOT SUPPORTED YET */
    zrpnquery    * type_101;
    zoctetstring * type_102;     /* NOT SUPPORTED YET */
  } u;
} zquery;



/*
 * Result set indicator
 */
#define Z_FALSE   (0)
#define Z_TRUE    (1)


/*
 * Z39.50 allows the client to assign the names to the result sets.
 * Every query can have each its own result set name.
 * Unfortunately servers can provide support for named result set or not.
 * If a server does not support named result sets, then a client can have
 * only one result set active at one time.
 *
 * Using version 3 of the standard the named result set option can be
 * negotiated at initialization time. With the version 2 of the standard
 * the only change for a client to be aware for support of named result sets
 * is to probe the servers: if the queries fail the error codes indicate
 * the reason of failure.
 *
 * The result set name "default" can be used to name result sets
 * in Z-associations with servers that do not support named result sets.
 * This is the only result set that a client can refer to in this cases.
 */

#define Z_DEFAULT_RESULTSETNAME    "default"

typedef struct
{
  zreferenceid         * referenceid;               /* OPTIONAL */
  unsigned long          smallsetupperbound;
  unsigned long          largesetlowerbound;
  unsigned long          mediumsetpresentnumber;
  int                    replaceindicator;          /* Z_FALSE or Z_TRUE */
  zresultsetid         * resultsetname;
  zdatabasenames       * databasenames;
  zelementsetnames     * smallsetelementsetnames;   /* OPTIONAL */
  zelementsetnames     * mediumsetelementsetnames;  /* OPTIONAL */
  zobjectid            * preferredrecordsyntax;     /* OPTIONAL */
  zquery               * query;
  zotherinformation    * additionalsearchinfo;      /* OPTIONAL */
  zotherinformation    * otherinfo;                 /* OPTIONAL */
} zsearchrequest;



/*
 * Default diagnostic record
 */
#define Z_ADDINFO_V2ADDINFO_DEFAULTDIAGFORMAT  1
#define Z_ADDINFO_V3ADDINFO_DEFAULTDIAGFORMAT  2

typedef struct
{
  zobjectid * diagnosticsetid;
  int         condition;
  struct
  {
    int choice;
    union
    {
      zvisiblestring       * v2addinfo;
      zinternationalstring * v3addinfo;
    } u;
  } addinfo;
} zdefaultdiagformat;


/*
 * Surrogate diagnostic record
 */
#define Z_DEFAULTFORMAT_DIAGREC      1
#define Z_EXTERNALLYDEFINED_DIAGREC  2
typedef struct
{
  int choice;
  union
  {
    zdefaultdiagformat * defaultformat;
    zexternal          * externallydefined;
  } u;
} zdiagrec;



#define Z_EXTERNALLYTAGGED_FRAGMENTSYNTAX     1
#define Z_NOTEXTERNALLYTAGGED_FRAGMENTSYNTAX  2
typedef struct
{
  int choice;
  union
  {
    zexternal    * externallytagged;
    zoctetstring * notexternallytagged;
  } u;
} zfragmentsyntax;



#define Z_RETRIEVAL_RECORD       1
#define Z_SURROGATE_DIAGNOSTIC   2
#define Z_STARTING_FRAGMENT      3
#define Z_INTERMEDIATE_FRAGMENT  4
#define Z_FINAL_FRAGMENT         5

/*
 * Retrieval record definition
 */
typedef struct
{
  int choice;
  union
  {
    zexternal       * retrievalrecord;
    zdiagrec        * surrogatediagnostic;
    zfragmentsyntax * startingfragment;
    zfragmentsyntax * intermediatefragment;
    zfragmentsyntax * finalfragment;
  } u;
} zretrievalrecord;



/*
 * Name plus record
 */
typedef struct
{
  zdatabasename    * name;         /* OPTIONAL */
  zretrievalrecord * record;
} znameplusrecord;


typedef struct
{
  int                size;
  znameplusrecord ** table;       /* pointer to the table of pointers */
} zresponserecords;



typedef struct
{
  int              size;
  zdiagrec      ** table;         /* pointer to the table of pointers */
} zmultiplenonsurdiagnostics;


/*
 * Response records to be included in the Search and Present response APDUs
 * may include both retrieval records and surrogate diagnostic records.
 *
 * A surrogate diagnostic record is included for each retrieval record
 * not accessible at the time of the request for the purpose of forming
 * a retrieval record.
 */

#define Z_RESPONSERECORDS_RECORDS            1
#define Z_NONSURROGATEDIAGNOSTIC_RECORDS     2
#define Z_MULTIPLENONSURDIAGNOSTICS_RECORDS  3

typedef struct
{
  int choice;
  union
  {
    zresponserecords           * responserecords;
    zdefaultdiagformat         * nonsurrogatediagnostic;
    zmultiplenonsurdiagnostics * multiplenonsurdiagnostics;
  } u;
} zrecords;


/*
 * result set status indicators.
 */
#define Z_SUBSET_RESULTSET   1
#define Z_INTERIM_RESULTSET  2
#define Z_NONE_RESULTSET     3


/*
 * status indicators (for both present and scan services)
 */
#define Z_SUCCESS_STATUS      1
#define Z_PARTIAL_1_STATUS    2
#define Z_PARTIAL_2_STATUS    3
#define Z_PARTIAL_3_STATUS    4
#define Z_PARTIAL_4_STATUS    5
#define Z_PARTIAL_5_STATUS    6                 /* only for scan */
#define Z_FAILURE_STATUS      7


typedef struct
{
  zreferenceid      * referenceid;              /* OPTIONAL */
  unsigned long       resultcount;
  unsigned long       numberofrecordsreturned;
  unsigned long       nextresultsetposition;
  int                 searchstatus;             /* Z_FALSE or Z_TRUE */
  int               * resultsetstatus;          /* OPTIONAL */
  int               * presentstatus;            /* OPTIONAL */
  zrecords          * records;                  /* OPTIONAL */
  zotherinformation * additionalsearchinfo;     /* OPTIONAL */
  zotherinformation * otherinfo;                /* OPTIONAL */
} zsearchresponse;



/*
 * -=-=-=-=-=-=-=-=-=-=-=-
 * -=  PRESENT SERVICE  =-
 * -=-=-=-=-=-=-=-=-=-=-=-
 */

/*
 * The origin may request additional ranges of records.
 * A range is a pair of N records beginning at position M
 */

typedef struct
{
  unsigned long startingposition;
  unsigned long numberofrecords;
} zrange;


typedef struct
{
  unsigned long    size;
  zrange        ** table;        /* pointer to the table of pointers */
} zaddranges;


#define Z_ELEMENTSETNAME_ELEMENTSPEC  1
#define Z_EXTERNALESPEC_ELEMENTSPEC   2
/*
 * Element specification
 */
typedef struct
{
  int choice;
  union
  {
    zinternationalstring * elementsetname;
    zexternal            * externalespec;
  } u;
} zelementspec;


typedef struct
{
  zobjectid    * schema;                   /* OPTIONAL */
  zelementspec * elementspec;              /* OPTIONAL */
} zspecification;


typedef struct
{
  zdatabasename  * db;
  zspecification * spec;
} zdbspecificelement;


typedef struct
{
  int size;
  zdbspecificelement ** table;    /* pointer to the table of pointers */
} zdbspecific;



typedef struct
{
  int          size;
  zobjectid ** table;             /* pointer to the table of pointers */
} zrecordsyntaxes;


typedef struct
{
  int               selectalternativesyntax;      /* Z_FALSE or Z_TRUE */
  zspecification  * generic;                      /* OPTIONAL */
  zdbspecific     * dbspecific;                   /* OPTIONAL */
  zrecordsyntaxes * recordsyntax;                 /* OPTIONAL */
} zcompspec;


#define Z_SIMPLE_RECORD_COMPOSITION   1
#define Z_COMPLEX_RECORD_COMPOSITION  2
typedef struct
{
  int choice;
  union
  {
    zelementsetnames * simple;
    zcompspec        * complex;
  } u;
} zrecordcomp;


typedef struct
{
  zreferenceid      * referenceid;               /* OPTIONAL */
  zresultsetid      * resultsetid;
  int                 resultsetstartpoint;
  int                 numberofrecordsrequested;
  zaddranges        * additionalranges;          /* OPTIONAL */
  zrecordcomp       * recordcomposition;         /* OPTIONAL */
  zobjectid         * preferredrecordsyntax;     /* OPTIONAL */
  unsigned long     * maxsegmentcount;           /* OPTIONAL */
  unsigned long     * maxrecordsize;             /* OPTIONAL */
  unsigned long     * maxsegmentsize;            /* OPTIONAL */
  zotherinformation * otherinfo;                 /* OPTIONAL */
} zpresentrequest;


typedef struct
{
  zreferenceid      * referenceid;               /* OPTIONAL */
  int                 numberofrecordsreturned;
  int                 nextresultsetposition;
  int                 presentstatus;
  zrecords          * records;                   /* OPTIONAL */
  zotherinformation * otherinfo;                 /* OPTIONAL */
} zpresentresponse;


/*
 * -=-=-=-=-=-=-=-=-=-=-=
 * -=  DELETE SERVICE  -=
 * -=-=-=-=-=-=-=-=-=-=-=
 */

typedef struct
{
  int             size;
  zresultsetid ** table;         /* pointer to the table of pointers */
} zresultsetlist;


/*
 * Z39.50 functions.
 */
#define Z_DELETE_LIST  1
#define Z_DELETE_ALL   2

typedef struct
{
  zreferenceid      * referenceid;                  /* OPTIONAL */
  int                 function;
  zresultsetlist    * resultsetlist;                /* OPTIONAL */
  zotherinformation * otherinfo;                    /* OPTIONAL */
} zdeleterequest;


#define Z_DELETE_SUCCESS                            1
#define Z_DELETE_RESULTSETDIDNOTEXIST               2
#define Z_DELETE_PREVIOUSLYDELETEDBYTARGET          3
#define Z_DELETE_SYSTEMPROBLEMATTARGET              4
#define Z_DELETE_ACCESSNOTALLOWED                   5
#define Z_DELETE_RESOURCECONTROLATORIGIN            6
#define Z_DELETE_RESOURCECONTROLATTARGET            7
#define Z_DELETE_BULKDELETENOTSUPPORTED             8
#define Z_DELETE_NOTALLRSLTSETSDELETEDONBULKDLTE    9
#define Z_DELETE_NOTALLREQUESTEDRESULTSETSDELETED  10
#define Z_DELETE_RESULTSETINUSE                    11


typedef struct
{
  zresultsetid * id;
  int            status;
} zliststatus;


typedef struct
{
  int            size;
  zliststatus ** table;
} zliststatuses;



typedef struct
{
  zreferenceid         * referenceid;               /* OPTIONAL */
  int                    status;
  zliststatuses        * liststatuses;              /* OPTIONAL */
  int                  * numbernotdeleted;          /* OPTIONAL */
  zliststatuses        * bulkstatuses;              /* OPTIONAL */
  zinternationalstring * message;                   /* OPTIONAL */
  zotherinformation    * otherinfo;                 /* OPTIONAL */
} zdeleteresponse;


/*
 * -=-=-=-=-=-=-=-=-=-=
 * -=  SCAN SERVICE  -=
 * -=-=-=-=-=-=-=-=-=-=
 */

typedef struct
{
  zreferenceid        * referenceid;                  /* OPTIONAL */
  zdatabasenames      * databasenames;
  zattributesetid     * attributeset;                 /* OPTIONAL */
  zattributesplusterm * termlistandstartpoint;
  int                 * stepsize;                     /* OPTIONAL */
  int                   numberoftermsrequested;
  int                 * preferredpositioninresponse;  /* OPTIONAL */
  zotherinformation   * otherinfo;                    /* OPTIONAL */
} zscanrequest;



typedef struct
{
  int                    size;
  zattributesplusterm ** table;       /* pointer to the table of pointers */
} zalternativeterms;


typedef struct
{
  char              * db;
  int               * num;              /* OPTIONAL */
  zotherinformation * otherdbinfo;      /* OPTIONAL */
} zbydatabase;


typedef struct
{
  int            size;
  zbydatabase ** table;
} zbydatabases;




#define Z_OCCURRENCES_GLOBAL      1
#define Z_OCCURRENCES_BYDATABASE  2
typedef struct
{
  int choice;
  union
  {
    int            global;
    zbydatabases * bydatabases;
  } u;
} zoccurrence;


typedef struct
{
  zattributelist    * attributes;
  zoccurrence       * occurrences;         /* OPTIONAL */
  zotherinformation * otheroccurinfo;      /* OPTIONAL */
} zoccurrencebyattribute;


typedef struct
{
  int                       size;
  zoccurrencebyattribute ** table;
} zoccurrencebyattributes;



typedef struct
{
  zterm                   * term;
  zinternationalstring    * displayterm;           /* OPTIONAL */
  zattributelist          * suggestedattributes;   /* OPTIONAL */
  zalternativeterms       * alternativeterm;       /* OPTIONAL */
  unsigned long           * globaloccurrences;     /* OPTIONAL */
  zoccurrencebyattributes * byattributes;          /* OPTIONAL */
  zotherinformation       * otherterminfo;         /* OPTIONAL */
} zterminfo;



#define Z_TERMINFO             1
#define Z_SURROGATEDIAGNOSTIC  2
typedef struct
{
  int choice;
  union
  {
    zterminfo * terminfo;
    zdiagrec  * surrogatediagnostic;
  } u;
} zentry;


typedef struct
{
  int       size;
  zentry ** table;                /* pointer to the table of pointers */
} zentries;


typedef struct
{
  int         size;
  zdiagrec ** table;              /* pointer to the table of pointers */
} znonsurrogatediagnostics;



/*
 * At least one of entries or nonsurrogatediagnostic must occur
 */
typedef struct
{
  zentries                 * entries;                  /* OPTIONAL */
  znonsurrogatediagnostics * nonsurrogatediagnostics;  /* OPTIONAL */
} zlistentries;


typedef struct
{
  zreferenceid      * referenceid;                  /* OPTIONAL */
  int               * stepsize;                     /* OPTIONAL */
  int                 scanstatus;
  unsigned long       numberofentriesreturned;
  unsigned long     * positionofterm;               /* OPTIONAL */
  zlistentries      * entries;                      /* OPTIONAL */
  zattributesetid   * attributeset;                 /* OPTIONAL */
  zotherinformation * otherinfo;                    /* OPTIONAL */
} zscanresponse;


typedef struct
{
  int                     size;
  zinternationalstring ** table; /* pointer to the table of pointers */
} zinputresultsetnames;


/*
 * Sort specification
 */
typedef struct
{
  zobjectid    * schema;              /* OPTIONAL */
  zelementspec * elementspec;         /* OPTIONAL */
} zsortspecification;


typedef struct
{
  zattributesetid * id;
  zattributelist  * list;
} zsortattributes;


/*
 * Sort key
 */
#define Z_SORTFIELD_SORTKEY       1
#define Z_ELEMENTSPEC_SORTKEY     2
#define Z_SORTATTRIBUTES_SORTKEY  3

typedef struct
{
  int choice;
  union
  {
    zinternationalstring * sortfield;
    zsortspecification   * elementspec;
    zsortattributes      * sortattributes;
  } u;
} zsortkey;



typedef struct
{
  zdatabasename * base;
  zsortkey      * key;
} zsortdbspec;


typedef struct
{
  int            size;
  zsortdbspec ** table;
} zdatbasespecific;


#define Z_SORT_GENERIC      1
#define Z_SORT_DB_SPECIFIC  2

/*
 * Sort element
 */
typedef struct
{
  int choice;
  union
  {
    zsortkey         * generic;
    zdatbasespecific * specific;
  } u;
} zsortelement;




#define Z_ABORT_MISSINGVALUEACTION             1
#define Z_NULL_MISSINGVALUEACTION              2
#define Z_MISSINGVALUEDATA__MISSINGVALUEACTION 3
typedef struct
{
  int choice;
  union
  {
    char           abort;
    char           null;
    zoctetstring * missingvaluedata;
  } u;
} zmissingvalueaction;


/*
 * Sort relation
 */
#define Z_ASCENDING_SORTRELATION             1
#define Z_DESCENDING_SORTRELATION            2
#define Z_ASCENDINGBYFREQUENCY_SORTRELATION  3
#define Z_DESCENDINGBYFREQUENCY_SORTRELATION 4

/*
 * Sort case sensitivity
 */
#define Z_CASESENSITIVE   1
#define Z_CASEINSENSITIVE 2
typedef struct
{
  zsortelement        * sortelement;
  int                   sortrelation;
  int                   casesensitivity;
  zmissingvalueaction * missingvalueaction;            /* OPTIONAL */
} zsortkeyspec;


typedef struct
{
  int             size;
  zsortkeyspec ** table;         /* pointer to the table of pointers */
} zsortsequence;


typedef struct
{
  zreferenceid         * referenceid;                  /* OPTIONAL */
  zinputresultsetnames * inputresultsetnames;
  zinternationalstring * sortedresultsetname;
  zsortsequence        * sortsequence;
  zotherinformation    * otherinfo;                    /* OPTIONAL */
} zsortrequest;


typedef struct
{
  int         size;
  zdiagrec ** table;
} zdiagnostics;


#define Z_EMPTY_RESULTSETSTATUS      1
#define Z_INTERIM_RESULTSETSTATUS    2
#define Z_UNCHANGED_RESULTSETSTATUS  3
#define Z_NONE_RESULTSETSTATUS       4
typedef struct
{
  zreferenceid      * referenceid;              /* OPTIONAL */
  int                 sortstatus;
  int               * resultsetstatus;          /* OPTIONAL */
  zdiagnostics      * diagnostics;              /* OPTIONAL */
  zotherinformation * otherinfo;                /* OPTIONAL */
} zsortresponse;



/*
 * -=-=-=-=-=-=-=-=-=-=-
 * -=  CLOSE SERVICE  =-
 * -=-=-=-=-=-=-=-=-=-=-
 */

/*
 * Warning: Close service may be used only when version 3 is in force
 */

#define Z_FINISHED           1
#define Z_SHUTDOWN           2
#define Z_SYSTEM_PROBLEM     3
#define Z_COST_LIMIT         4
#define Z_RESOURCES          5
#define Z_SECURITY_VIOLATION 6
#define Z_PROTOCOL_ERROR     7
#define Z_LACK_OF_ACTIVITY   8
#define Z_PEER_ABORT         9
#define Z_UNSPECIFIED        10

typedef struct
{
  zreferenceid         * referenceid;                 /* OPTIONAL */
  int                    reason;
  zinternationalstring * diagnostic;                  /* OPTIONAL */
  zresourcereportid    * resourcereportformat;        /* OPTIONAL */
  zexternal            * resourcereport;              /* OPTIONAL */
  zotherinformation    * otherinfo;                   /* OPTIONAL */
} zclose;



typedef struct
{
  union
  {
    zinitializerequest  * initrequest;
    zinitializeresponse * initresponse;
    zsearchrequest      * searchrequest;
    zsearchresponse     * searchresponse;
    zpresentrequest     * presentrequest;
    zpresentresponse    * presentresponse;
    zdeleterequest      * deleterequest;
    zdeleteresponse     * deleteresponse;
    void                * accesscontrolrequest;     /* Not Yet Implemented */
    void                * accesscontrolresponse;    /* Not Yet Implemented */
    void                * resourcecontrolrequest;   /* Not Yet Implemented */
    void                * resourcecontrolresponse;  /* Not Yet Implemented */
    void                * triggerrequest;           /* Not Yet Implemented */
    void                * resourcereportrequest;    /* Not Yet Implemented */
    void                * resourcereportresponse;   /* Not Yet Implemented */
    zscanrequest        * scanrequest;
    zscanresponse       * scanresponse;
    zsortrequest        * sortrequest;
    zsortresponse       * sortresponse;
    void                * segmentrequest;           /* Not Yet Implemented */
    void                * extendedservicesrequest;  /* Not Yet Implemented */
    void                * extendedservicesresponse; /* Not Yet Implemented */
    zclose              * close;
  } u;
} zpdu;



/*
 * ANSI C Prototypes for the API routines.
 */

#if defined(WIN32)
# include <windows.h>
# if defined(LINK_CYGNUSLIB)
#  define DECEXP
#  define EXPORT_FUNC
# else
#  define DECEXP __declspec(dllexport)
#  define EXPORT_FUNC _stdcall
# endif /* LINK_CYGNUSLIB */
# undef WINAPI
# define WINAPI _stdcall
#else
# if defined(_WINDOWS)
#  include <windows.h>
#  define EXPORT_FUNC WINAPI
# else
#  define DECEXP
#  define EXPORT_FUNC
#  define WINAPI
# endif /* ! _WINDOWS */
#endif /* ! WIN32 */



DECEXP int    EXPORT_FUNC z_accept     (int fd, int newfd);
DECEXP int    EXPORT_FUNC z_bind       (int fd, char * address, int role);
DECEXP int    EXPORT_FUNC z_close      (int fd);
DECEXP int    EXPORT_FUNC z_connect    (int fd, char * address);
DECEXP char * EXPORT_FUNC z_error      (void);
DECEXP int    EXPORT_FUNC z_free       (int ztype, zpdu * pdu);
DECEXP int    EXPORT_FUNC z_listen     (int fd);
DECEXP int    EXPORT_FUNC z_look       (int fd);
DECEXP int    EXPORT_FUNC z_open       (const char * provider, int oflags);
DECEXP int    EXPORT_FUNC z_rcv        (int fd, int * ztype, zpdu * pdu, int * flags);
DECEXP int    EXPORT_FUNC z_rcvconnect (int fd);
DECEXP int    EXPORT_FUNC z_snd        (int fd, int ztype, zpdu * pdu, int flags);
DECEXP int    EXPORT_FUNC z_state      (int fd);
DECEXP void   EXPORT_FUNC z_version    (void);

#if defined(_WINDOWS) || defined(WIN32)
DECEXP long   EXPORT_FUNC z_winnofile  (void);
DECEXP int    EXPORT_FUNC z_winerrno   (void);
#endif


#if defined(ZAPI_TRACE)

/*
 * Define types, constants, macros and function declarations
 * (in this order) for all tracing interaction at User level.
 */

extern int    z_trace;
extern FILE * z_fd;

/*
 * Bit mask values to control tracing at API level.
 */
#define Z_NONE_BIT          (0x000000)
#define Z_ENTER_BIT         (0x000001)
#define Z_INPUT_BIT         (0x000002)
#define Z_OUTPUT_BIT        (0x000004)
#define Z_EXIT_BIT          (0x000008)
#define Z_TIME_BIT          (0x000010)
#define Z_PID_BIT           (0x000020)
#define Z_RETURN_BIT        (0x000040)
#define Z_ERROR_BIT         (0x000080)
#define Z_BUFFER_BIT        (0x000100)
#define Z_PDU_ONLY_BIT      (0x000200)
#define Z_IN_PDUS_BIT       (0x000400)
#define Z_OUT_PDUS_BIT      (0x000800)
#define Z_WRONG_PDUS_BIT    (0x001000)
#define Z_BYTES_BIT         (0x002000)
#define Z_ALL_BITS          (0xffffff)


/*
 * Default values for the tracing variables.
 */
#define DEFAULT_Z_TRACE \
 (Z_ENTER_BIT  | Z_EXIT_BIT    | Z_TIME_BIT  | \
  Z_PID_BIT    | Z_RETURN_BIT  | Z_ERROR_BIT | \
  Z_BUFFER_BIT | Z_IN_PDUS_BIT | Z_OUT_PDUS_BIT)

#define DEFAULT_Z_FD       stdout
#define DEFAULT_Z_LOGFILE  "zeta.log"


#endif /* ZAPI_TRACE */
