#                                                                -*- perl -*-
# 0.generate.t - an automatic test scripts generator
#
# ZETA Perl - Perl binding for the ANSI/NISO Z39.50-1995 Protocol
# Copyright (c) 1996, 1999 Finsiel S.p.A.
#
# Jama Musse Jama <jama@tecsiel.it> 4Q 1998
# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

use ZETA;

# common subroutines
require "common.pl";

sub SetDefaults
  {
    # default values of global variables
    $target           = "www.promemoria.net";
    $port             = "2222";
    $database         = "test";
    $resultset        = "zeta";
    $query            = "1=1016 legno";
    $attrsetid        = "1.2.840.10003.3.1";
    $setname          = "brief";
    $syntax           = "1.2.840.10003.5.101";
    $term             = "1=4 a";
    $criteria         = "1=4";
    $user             = "";
    $password         = "";
    $attind           = 0;            # default is BIB1
    $recind           = 19;           # default is GRS1
  }

sub WriteToConfigFile
  {
    
    open (OUTPUT, ">$FILERC") || die "Cannot create $FILERC configuration file";
    print OUTPUT $HEAD_CONFIGFILE;

    print OUTPUT "#\n";
    print OUTPUT "# You need to provide a reachable z39.50 target.\n";
    print OUTPUT "# Its name can be provided as a valid hostname or its IP address.\n";
    print OUTPUT "#\n";
    print OUTPUT "\$target = \"$target\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# Give here the port number of the target\n";
    print OUTPUT "#\n";
    print OUTPUT "\$port = \"$port\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# Database name\n";
    print OUTPUT "#\n";
    print OUTPUT "\$database = \"$database\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# Resultset name\n";
    print OUTPUT "#\n";
    print OUTPUT "\$resultset = \"$resultset\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# You need to provide a simple query in order to test\n";
    print OUTPUT "# the search and present methods.\n";
    print OUTPUT "# Your query must create a non empty result set.\n";
    print OUTPUT "#\n";
    print OUTPUT "\$query = \"$query\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# You need to specify an attribute set id supported by your target\n";
    print OUTPUT "# The ZETA Perl module supports both the string and dot notation.\n";
    print OUTPUT "#\n";
    print OUTPUT "# Appendix 3 of ANSI/NISO Z39.50-1995 registers the following:\n";
    print OUTPUT "#\n";
    $nI = 0;
    while ($nI <= $#attsets)
      {
	print OUTPUT "# $attsets[$nI][0]      \"$attsets[$nI][1]\"\n";
	$nI ++;
      }
    print OUTPUT "\$attind = $attind;\n";
    print OUTPUT "\$attrsetid = \"$attsets[$attind][1]\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# Which element set name has to be used for the presentation of records.\n";
    print OUTPUT "#\n";
    print OUTPUT "\$setname = \"$setname\";\n\n";

    print OUTPUT "# You need to specify a record syntax for the presentation of records\n";
    print OUTPUT "#\n";
    print OUTPUT "# Appendix 5 of ANSI/NISO Z39.50-1995 defines the following\n";
    print OUTPUT "# object identifiers for use with the record syntaxes:\n";
    print OUTPUT "#\n";
    $nI = 0;
    while ($nI <= $#recsyn)
      {
	print OUTPUT "# $recsyn[$nI][0]  \t\"$recsyn[$nI][1]\"\n";
	$nI ++;
      }
    print OUTPUT "\$recind = $recind;\n";
    print OUTPUT "\$syntax = \"$recsyn[$recind][1]\";\n\n";
    print OUTPUT "#\n";
    print OUTPUT "# You need to provide a simple term in order to test the scan method.\n";
    print OUTPUT "#\n";
    print OUTPUT "\$term =\"$term\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# If your target supports the sorting method, you need\n";
    print OUTPUT "# to provide which sorting criteria you want to\n";
    print OUTPUT "# apply your output. [see example]\n";
    print OUTPUT "#\n";
    print OUTPUT "\$criteria = \"$criteria\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# If your target applies some authentication policy with name and passwd\n";
    print OUTPUT "# you need to provide valid username and password\n";
    print OUTPUT "#\n";
    print OUTPUT "\$user = \"$user\";\n";
    print OUTPUT "\$password = \"$password\";\n\n";

    print OUTPUT "#\n";
    print OUTPUT "# Other global variables\n";
    print OUTPUT "#\n";
    print OUTPUT "\$Verbose = 0;\n\n1;";
    close OUTPUT;
  }

sub InitGlobalVars
  {
    my $tmpstr;
    my $ReWriteConfig = 0;

    print STDERR "\nEnter target name .......... [$target]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$target = $tmpstr;
	$ReWriteConfig = 1;
      }

    print STDERR "Enter port number .......... [$port]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$port = $tmpstr;
	$ReWriteConfig = 1;
      }

    print STDERR "Enter database name ........ [$database]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$database = $tmpstr;
	$ReWriteConfig = 1;
      }
    print STDERR "Choose a result set name ... [$resultset]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$resultset = $tmpstr;
	$ReWriteConfig = 1;
      }

    print STDERR "Choose an attribute set id .\n";
    print STDERR "\tAppendix 3 of ANSI/NISO Z39.50-1995 registers the following:\n";
    $nI = 0;
    while ($nI <= $#attsets)
      {
	print STDERR "\t[$nI: $attsets[$nI][0]] ";
	$nI ++;
	print STDERR "\n" unless ($nI%4);
      }
    print STDERR "\n\tChoose the number of your preferred attribute set [$attind]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr && $tmpstr <= $#attsets)
      {
	$attind = $tmpstr;
	$attrsetid = $attsets[$tmpstr][1];
	$ReWriteConfig = 1;
      }

    print STDERR "Query ...................... [$query]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$query = $tmpstr;
	$ReWriteConfig = 1;
      }
    
    print STDERR "Element set name ........... [$setname]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$setname = $tmpstr;
	$ReWriteConfig = 1;
      }
    
    print STDERR "Record syntax .............. \n";
    $nI = 0;
    while ($nI <= $#recsyn)
      {
	print STDERR "\t[$nI: $recsyn[$nI][0]]";
	$nI ++;
	print STDERR "\n" unless ($nI%4);
      }
    print STDERR "\n\tChoose the number of your preferred record syntax [$recind]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr && $tmpstr <= $#recsyn)
      {
        $recind = $tmpstr;
	$syntax = $recsyn[$recind][1];
	$ReWriteConfig = 1;
      }
    
    print STDERR "Term to scan ............... [$term]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$term = $tmpstr;
	$ReWriteConfig = 1;
      }
    
    print STDERR "Sort criteria .............. [$criteria]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$criteria = $tmpstr;
	$ReWriteConfig = 1;
      }

    print STDERR "Enter user name ............ [$user]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$user = $tmpstr;
	$ReWriteConfig = 1;
      }

    print STDERR "Password ................... [$password]\> ";
    chop ($tmpstr = <STDIN>);
    if ($tmpstr)
      {
	$password = $tmpstr;
	$ReWriteConfig = 1;
      }

    if ($ReWriteConfig)
      {
	unlink $FILERC;
	&WriteToConfigFile ();
      }

  }


sub AskIfIChange ()
  {
    my $tmpstr;
    my $ok       = 0;

    print STDERR "\nThe following parameters are currently being used\n\n";

    print STDERR "\tTarget name .......... [$target]\>\n";
    print STDERR "\tPort number .......... [$port]\>\n";
    print STDERR "\tDatabase name ........ [$database]\>\n";
    print STDERR "\tResult set name ...... [$resultset]\>\n";
    print STDERR "\tQuery ................ [$query]\>\n";
    print STDERR "\tAttribute set ........ [$attsets[$attind][0] ($attrsetid)]\>\n";
    print STDERR "\tElement set name ..... [$setname]\>\n";
    print STDERR "\tRecord syntax ........ [$recsyn[$recind][0] ($syntax)]\>\n";
    print STDERR "\tTerm to scan ......... [$term]\>\n";
    print STDERR "\tSort criteria ........ [$criteria]\>\n";
    print STDERR "\tEnter user name ...... [$user]\>\n";
    print STDERR "\tPassword ............. [$password]\>\n\n";

    print STDERR "Do you want to change these values [y/N/d/h]\> ";
    chop ($tmpstr = <STDIN>);
    while (! $ok)
      {
	if (! $tmpstr || lc($tmpstr) eq 'n')
	  {
	    $ok = 1;
	  }
	elsif (lc($tmpstr) eq 'y')
	  {
	    $ChangeValues = 1;
	    $ok = 1;
	  }
	elsif (lc($tmpstr) eq 'd')
	  {
	    &SetDefaults ();
	    &WriteToConfigFile ();
	    $ok = 1;
	  }
	else
	  {
	    print STDERR "choose one of: \n\t[y] to change the above listed values\n\t[n] not change values (DEFAULT)\n\t[d] reset to the System Wide default values\n\t[h] display this information\nInsert [y/N/d/h] ";
	  }
	if (! $ok)
	  {
	    chop ($tmpstr = <STDIN>);
	  }

      }

    return $ChangeValues;

  }


sub GenerateTests()
  {
    print STDERR "ZETA Perl module test application: Version: $ZETA::VERSION\n";
    print STDERR "
   In order to test the functionality of this module, you need to have
   a reachable Z39.50 target.  Please choose a target name, the Port number,
   database name, the profile type, and type-1 query string.  In case your
   target applies some authentication policy, you may need to choose also
   a valid user name and password.  If you are not sure what you are doing,
   or if you do not have access to a Z39.50 target, please leave these
   parameters to their respective default values.\n\n";

    # Check if the config file exists, and in case create a default one
    if (! -e $FILERC)
      {
	&WriteToConfigFile ();
      }
    # global variables amd constansts are found here
    require "$FILERC";

    # inform the user about the ZETA target for testing
    if (&AskIfIChange ())
      {
	# give the user the possibility to modify the ZETA target information
	&InitGlobalVars ();
      }

    # return any way "test ok"
    SayOk ($testno, 1, 0);

  }

BEGIN
  {

    # more constants
    $HEAD_CONFIGFILE  = "#                                                                -*- text -*-
# zetaperlrc - System Wide default configuration file for ZETA Perl Module
#
# ZETA Perl - Perl binding for the ANSI/NISO Z39.50-1995 Protocol
# Copyright (c) 1996, 1999 Finsiel S.p.A.
#
# Rocco Carbone <rocco\@tecsiel.it>
# Jama Musse Jama <jama\@tecsiel.it> 4Q 1998
#
# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
# The purpose of this file is to provide the necessary information
# for the ZETA perl module to test all its methods and functionalities.
# If this file is deleted, it will be generated automatically
# with make test.
#
# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#\n\n";
      
    @attsets = (
		['BIB1' , '1.2.840.10003.3.1'],
		['EXP1' , '1.2.840.10003.3.2'],
		['EXT1' , '1.2.840.10003.3.3'],
		['CCL1' , '1.2.840.10003.3.4'],
		['GILS' , '1.2.840.10003.3.5'],
		['STAS' , '1.2.840.10003.3.6'],
		['CIMI1', '1.2.840.10003.3.8']
	       );
    
    @recsyn = (
	       ['UNIMARC'      , '1.2.840.10003.5.1'],
	       ['INTERMARC'    , '1.2.840.10003.5.2'],
	       ['CCF'          , '1.2.840.10003.5.3'],
	       ['USMARC'       , '1.2.840.10003.5.10'],
	       ['UKMARC'       , '1.2.840.10003.5.11'],
	       ['NORMARC'      , '1.2.840.10003.5.12'],
	       ['LIBRISMARC'   , '1.2.840.10003.5.13'],
	       ['DANMARC'      , '1.2.840.10003.5.14'],
	       ['FINMARC'      , '1.2.840.10003.5.15'],
	       ['MAB'          , '1.2.840.10003.5.16'],
	       ['CANMARC'      , '1.2.840.10003.5.17'],
	       ['SBN'          , '1.2.840.10003.5.18'],
	       ['PICAMARC'     , '1.2.840.10003.5.19'],
	       ['AUSMARC'      , '1.2.840.10003.5.20'],
	       ['IBERMARC'     , '1.2.840.10003.5.21'],
	       ['EXPLAIN'      , '1.2.840.10003.5.100'],
	       ['SUTRS'        , '1.2.840.10003.5.101'],
	       ['OPAC'         , '1.2.840.10003.5.102'],
	       ['SUMMARY'      , '1.2.840.10003.5.103'],
	       ['GRS1'         , '1.2.840.10003.5.105'],
	       ['EXTSERVICES'  , '1.2.840.10003.5.106']
	      );
    
    
    &SetDefaults();
    $nI = 0;
  }

print "1..1\n";

$testno = 0;
&GenerateTests ();
