#                                                                -*- perl -*-
# 7.scan.t - test of ZETA Perl Module for Scan service
#
# ZETA Perl - Perl binding for the ANSI/NISO Z39.50-1995 Protocol
# Copyright (c) 1996, 1999 Finsiel S.p.A.
#
# Jama Musse Jama <jama@tecsiel.it> 4Q 1998
# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

use ZETA;
use Benchmark;

# common subroutines
require "common.pl";

# global variables amd constansts are found here
require "$FILERC";

sub TestScan ()
  {
    # local variables
    my $zc;
    my $status;
    my $reason;
    my $result;
    my $refid = 1;
    my $newrefid;
    my $close = 1;
    my $found;
    my @entries;
    my $items = 1;
    my $position = 1;

    my $t1;
    my $td;
    my $tds;

    print STDERR "\n\t<<-- Testing scan service\n" if ($Verbose);

    $t1 = new Benchmark;

    #
    # establish a connection with a Z39.50 Target over TCP/IP
    #
    $zc = ZETA::Connection::New ($target, $port);

    $td  = Benchmark::timediff (Benchmark->new, $t1);
    $tds = Benchmark::timestr ($td);

    if (! $zc)
      {
	print STDERR "\t<<-- Connection to $target:$port failed\n" if ($Verbose);
	&TestAbort ();
      }
    else
      {
	&SayOk ($testno, $zc, $tds);

	print STDERR "\t<<-- Initializing the association\n" if ($Verbose);

	$t1 = new Benchmark;

	#
	# initialize the association
	#
	($status, $reason, $result, $newrefid) =
	  $zc->Init ($user, $password, $refid);

	my $td  = Benchmark::timediff (Benchmark->new, $t1);
	my $tds = Benchmark::timestr ($td);

	if (! $status)
	  {
	    print STDERR "\t<<-- Init for user [$user/$password with reference id = $refid] failed due to $reason\n" if ($Verbose);
	    &TestAbort ();
	  }
	elsif (! $result)
	  {
	    print STDERR "\t<<-- Target rejected the request\n" if ($Verbose);
	    &TestAbort ();
	  }
	else
	  {
	    print STDERR "\t<<-- Target initialized\n" if ($Verbose);
	  }

	&SayOk ($testno, $status, $tds);

	print STDERR "\t<<-- Trying to browse $database for $term\n" if ($Verbose);

	#
	# try to create a result set
	#
	$t1 = new Benchmark;

	($status, $reason, $found, @entries) =
	  $zc->Scan ($database, $term, $items, $position);

	$td = Benchmark::timediff (Benchmark->new, $t1);
	$tds= Benchmark::timestr ($td);

	if (! $status)
	  {
	    print STDERR "\t<<-- Scan failed due to $reason\n" if ($Verbose);
	    &TestAbort ();
	  }
	else
	  {
	    if ($Verbose)
	      {
		print STDERR "\t<<-- $found entries found\n";
		foreach (@entries)
		  {
		    $a = chomp ($_);
		    print STDERR "\t<<-- ($a)\n";
		  }
	      }
	  }

	SayOk ($testno, $status, $tds);

	print STDERR "\t<<-- Closing the connection [$close]\n" if ($Verbose);

	$t1 = new Benchmark;

	#
	# close the association
	#
	($status, $reason) = $zc->Close ($close);

	$td  = Benchmark::timediff (Benchmark->new, $t1);
	$tds = Benchmark::timestr ($td);

	if (! $status)
	  {
	    print STDERR "\t<<-- Close failed due to $reason\n" if ($Verbose);
	    &TestAbort ();
	  }
	else
	  {
	    print STDERR "\t<<-- Association closed\n" if ($Verbose);
	  }

	&SayOk ($testno, $status, $tds);
      }
}

# here is the main
BEGIN
  {
    ;
  }

print "1..4\n";

if (-e "../$FILERC")
  {
    print STDERR "Configuration file missing\n";
  }

$testno = 0;
&TestScan ();
