#
# test.pl - a template test script for the ZETA Perl
#
# ZETA Perl - Perl binding for the ANSI/NISO Z39.50-1995 Protocol
# Copyright (c) 1996, 1999 Finsiel S.p.A.
#
# Rocco Carbone <rocco@tecsiel.it> 4Q 1996
# Jama Musse Jama <jama@tecsiel.it> 3Q 1998
# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#

use ZETA;

# default values

$host      = 'www.promemoria.net';
$port      = 2222;
$id        = '';
$database  = 'test';
$resultset = "Y2K";
$attrsetid = 'bib1';
$query     = "1=1016 madonna \@or\@ 1=1016 legno";
$howmany   = 3;
$setname   = 'brief';
$syntax    = 'grs1';

$scanterm  = '1=4 "a"';
$scanitems = 10;

$criteria  = '1=4';


$you = getlogin ();
print "\t\nNice to meet you >>> $you <<<\n";

zetaversion();

print "\n";


$| = 1;  print "enter z39.50 target name:    [$host] "; $| = 0;
chomp ($choice = <STDIN>);
$host = $choice if ($choice);

$| = 1;  print "enter z39.50 target port:    [$port] "; $| = 0;
chomp ($choice = <STDIN>);
$port = $choice if ($choice);

$| = 1;  print "enter z39.50 database name:  [$database] "; $| = 0;
chomp ($choice = <STDIN>);
$database = $choice if ($choice);

$| = 1;  print "enter z39.50 attribute id:   [$attrsetid] "; $| = 0;
chomp ($choice = <STDIN>);
$database = $choice if ($choice);

$| = 1;  print "enter z39.50 type 1 query:   [$query] "; $| = 0;
chomp ($choice = <STDIN>);
$query = $choice if ($choice);

$| = 1;  print "enter how many hits to show: [$howmany] "; $| = 0;
chomp ($choice = <STDIN>);
$howmany = $choice if ($choice);

$| = 1;  print "enter z39.50 record syntax:  [$syntax] "; $| = 0;
chomp ($choice = <STDIN>);
$syntax = $choice if ($choice);



#
# create a connection with a z39.50 Target
#
print "\n";
print "connecting to $host:$port ... ";

$connection = ZETA::Connection::New ($host, $port);
if (! $connection)
  {
    die "failed $host:$port\n";
  }
else
  {
    print "ok\n\n";
  }

#
# initialize the association
#
print "initializing ... ";

($status, $reason, $result, $reference) =
    $connection->Init ('', '', $id);
if (! $status)
  {
    die "Init had a failure --- $reason\n";
  }
else
  {
    print "ok\n\n";
    print "Your identifier is $reference.\n" if ($reference); 
  }


#
# create a result set
#
print "searching database $database for $query ... ";

($status, $reason, $hits, $info) =
  $connection->Search ($database, $resultset, $query, $attrsetid);

if (! $status)
  {
    die "Search had a failure --- $reason\n";
  }
else
  {
    print "ok\n\n";
    print "-->> result set $resultset created. it contains references to $hits records\n\n";
  }

#
# show the records of the result set
#


$start = 1;
for $record (1 .. $hits)
  {
    print "retrieving $howmany records of $resultset according to $syntax ... " if ($hits);

    ($status, $reason, $returned, @records) =
      $connection->Present ($resultset, $howmany, $start, $setname, $syntax);
    print "ok\n";
    if (! $status)
      {
	die "Present had a failure --- $reason\n";
      }
    else
      {
        $i = $start;
	print "-->> $returned records returned\n\n\n";

	foreach (@records)
	  {
	    $a = $_;
	    print ("--- BEGIN $i of $hits ---\n\n");
	    print $a,"\n\n";
	    print ("---  END  $i of $hits ---\n\n");
	    $| = 1;  print "more? (y/n) "; $| = 0;
	    chomp ($more = <STDIN>);
	    if ($more eq 'n')
	      {
		last;
	      }
	    $i ++;
	  }
      }
    $start += $returned;

    if ($more eq 'n')
      {
	last;
      }
  }


#
# scan the database
#
print "scanning database $database for $scanterm ... ";

($status, $reason, $found, @entries) =
  $connection->Scan ($database, $scanterm, $scanitems, 1);

if (! $status)
{
   die "Scan had a failure --- $reason\n";
}
else
{
  print "ok\n\n";
  print "-->> $found entries found\n";
  for (@entries)
  {
    print "      $_\n";
  }
}


print "\n";

$| = 1;  print "enter z39.50 sorting criteria:  [$criteria] "; $| = 0;
chomp ($choice = <STDIN>);
$criteria = $choice if ($choice);


#
# sort the result set
#
print "Trying to sort $resultset ...\n";

print "sorting result set $resultset accorging to $criteria ... ";

($status, $reason) = $connection->Sort ($resultset, $criteria);

if (! $status)
  {
    die "Sort had a failure --- $reason\n";
  }
else
  {
    print "ok\n\n";
  }
