/*
 * zutils.c - common utilities for ZETA Perl
 *
 * ZETA Perl - Perl binding for the ANSI/NISO Z39.50-1995 Protocol
 * Copyright (c) 1996, 1999 Finsiel S.p.A.
 *
 * Rocco Carbone <rocco@tecsiel.it> 4Q 1996
 * Maria Grazia Toro <toro@tecsiel.it> 1Q 1998
 * -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */



/*
 * operating system header file(s)
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * ZETA header file(s)
 */
#include "zeta.h"
#include "grs1.h"

char * z_diagbib12a (int code);
static int grs1len (grs1_genericrecord * g);
static int create_grs1_record (grs1_genericrecord* d, char *outf);
static char *grs1treestr = NULL;
static int grs1indent = 0;
static char * buildsurrogate (zdiagrec *d, char * name);
static char * builddefaultdiag (zdefaultdiagformat *d, char * name);
char * builddiagnostics (unsigned long num, zdiagrec ** table);
zattributeelement * newattr (int type, int value, zattributesetid * id);


/*
 * Add the element pointed to by 'r' to the table of result set
 */
void addtotableofresultset (zresultsetlist * rstable, zresultsetid * r)
{
  rstable -> table = realloc (rstable -> table, (rstable -> size + 1) *
			      sizeof (zresultsetlist *));
  rstable -> table [rstable -> size] = malloc (strlen (r) + 1);
  strcpy (rstable -> table [rstable -> size ++], r);
}



/*
 * Add the element pointed to by 'd' to the table of databases
 */
void addtotableofdatabases (zdatabasenames * dtable, char * d)
{
  int i, j, k;
  char * p;
  for(i = 0, j = 1; i < strlen (d); i ++)
    if (d[i] == '@')
      j ++;
  dtable -> table =
    realloc (dtable -> table,
	     (dtable -> size + j) * sizeof (zdatabasenames *));
  for(i = 0; j > 1; i ++)
    {
      if (d[i] == '@')
        {
	  p = &(d[i+1]);
	  j --;
         d [i] = '\0';
         dtable -> table [dtable -> size] = malloc (strlen (d) + 1);
         strcpy (dtable -> table [dtable -> size], d);
         dtable -> size ++;
         d = p;
	 i = -1;
        }
    }
  dtable -> table [dtable -> size] = malloc (strlen (d) + 1);
  strcpy (dtable -> table [dtable -> size], d);
  dtable -> size ++;
}


/*
 * Append 'e' to the table of attributes
 */
void addtolistofattr (zattributelist * atable, zattributeelement * e)
{
  if (! e)
    return;

  atable -> table =
    realloc (atable -> table,
	     (atable -> size + 1) * sizeof (zattributeelement *));
  if (! atable -> table)
    return;

  atable -> table [atable -> size] = e;
  atable -> size ++;
}



#define ISO2709_RS 035
#define ISO2709_FS 036
#define ISO2709_IDFS 037


static int atoi_n (char * buf, int len)
{
  int val = 0;

  while (-- len >= 0)
    {
      if (isdigit (* buf))
	val = val * 10 + (* buf - '0');
      buf ++;
    }
  return (val);
}


/*
 * GRS.1 Record len
 */
static int  grs1datalen (grs1_data * d, int tagtype, int tagval,
			 char * tagstr, int tagocc)
{
  int ret = 0;
  char tmp [100];
  int ret1;

  if (! d)
    return (ret);

  if (tagstr && (tagocc != (-1)))
    ret1 = sprintf (tmp,"[%d,%s, %d]", tagtype, tagstr, tagocc);
  else if (tagstr)
    ret1 = sprintf (tmp ,"[%d,%s]", tagtype, tagstr);
  else if (tagocc != (-1))
    ret1 = sprintf (tmp,"[%d,%d, %d]", tagtype, tagval, tagocc);
  else
    ret1 = sprintf (tmp ,"[%d,%d]", tagtype, tagval);
  if (d -> choice != GRS1_SUBTREE)
    ret += ret1 + grs1indent;
  switch (d -> choice)
    {

    case GRS1_NUMERIC:
      ret += sprintf (tmp," : %d\n", d -> u . numeric);
      break;
    case GRS1_DATE:
      ret += strlen (d -> u . date) + 4;
      break;
    case GRS1_STRING:
      ret += strlen (d -> u . string) + 4;
      break;
    case GRS1_OID:
      ret += strlen (d -> u . oid) + 4;
      break;
    case GRS1_EXT:
    case GRS1_OCTETS:
    case GRS1_TRUEORFALSE:
    case GRS1_INTUNIT:
    case GRS1_ELEMENTNOTTHERE:
    case GRS1_ELEMENTEMPTY:
    case GRS1_NODATAREQUESTED:
    case GRS1_DIAGNOSTIC:
      ret += 4;
      break;
    case GRS1_SUBTREE:
      grs1indent += ret1;
      ret += grs1len ((grs1_genericrecord *) d -> u . subtree);
      grs1indent -= ret1;
      break;
    }

  return (ret);
}


static int grs1len (grs1_genericrecord * g)
{
  int i;
  int ret = 0;
  int tagval, tagset, tagocc;
  char *tagstr;

  if (! g || ! g -> tree)
    return (ret);

  for (i = 0; i < g -> size; i ++)
    {
      if (! g -> tree [i])
	continue;

      /* type */
      if (g -> tree [i] -> type)
	tagset = * g -> tree [i] -> type;
      else
	tagset = 0;
      /* value */
      if (g -> tree [i] -> value)
	{
	  if (g -> tree [i] -> value -> choice == Z_STRING_STRINGORNUMERIC)
	    {
	      tagstr = g -> tree [i] -> value -> u . string;
	      tagval = 0;
	    }
	  else if (g -> tree [i] -> value -> choice == Z_NUMBER_STRINGORNUMERIC)
	    {
	      tagval = g -> tree [i] -> value -> u . numeric;
	      tagstr = NULL;
	    }
	}

      /* occurrence */
      if (g -> tree [i] -> occurrence)
	tagocc = * g -> tree [i] -> occurrence;
      else
	tagocc = -1;

      ret += grs1datalen (g -> tree [i] -> content, tagset, tagval,
			  tagstr, tagocc);
    }
  return(ret);
}


/*
 * create GRS.1 data
 */
static int create_grs1_data_record (grs1_data * d, char * outf, int
				    tagtype, int tagval, char *tagstr,
				    int tagocc)
{
  int ret = 0;
  char tmp[100];
  int ret1;
  char *ptr;

  if (! d)
    return (ret);

  if (tagstr && tagocc != -1)
    ret1 = sprintf (tmp, "[%d,%s, %d]", tagtype, tagstr, tagocc);
  else if (tagstr)
    ret1 = sprintf (tmp, "[%d,%s]", tagtype, tagstr);
  else if (tagocc != -1)
    ret1 = sprintf (tmp, "[%d,%d, %d]", tagtype, tagval, tagocc);
  else
    ret1 = sprintf (tmp, "[%d,%d]", tagtype, tagval);
  if (d -> choice != GRS1_SUBTREE)
    {
      if (grs1indent)
	{
	  outf [0] = '\0';
	  outf = strcat (outf, grs1treestr);
	  ret = grs1indent;
	}
      else
	outf [0] = '\0';
      ptr = strcat (& outf [ret], tmp);
      ret += ret1;
    }

  switch (d -> choice)
    {
    case GRS1_NUMERIC:
      ret += sprintf (& outf [ret], " : %d\n", d -> u . numeric);
      break;
    case GRS1_DATE:
      ret += sprintf (& outf [ret], " : ");
      if (d -> u . date)
	memcpy (& outf [ret], d -> u . date, strlen (d -> u . date));
      ret += strlen (d -> u . date);
      outf [ret ++] = '\n';
      break;
    case GRS1_STRING:
      ret += sprintf (& outf [ret], " : ");
      if (d -> u . string)
	memcpy (& outf [ret], d -> u . string, strlen (d -> u . string));
      ret += strlen (d -> u . string);
      outf [ret ++] = '\n';
      break;
    case GRS1_OID:
      ret +=sprintf (& outf [ret], " : ");
      if (d -> u . oid)
	memcpy (& outf [ret], d -> u . oid, strlen (d -> u . oid));
      ret += strlen (d -> u . oid);
      outf [ret ++] = '\n';
      break;
    case GRS1_EXT:
    case GRS1_TRUEORFALSE:
    case GRS1_OCTETS:
    case GRS1_INTUNIT:
    case GRS1_ELEMENTNOTTHERE:
    case GRS1_ELEMENTEMPTY:
    case GRS1_NODATAREQUESTED:
    case GRS1_DIAGNOSTIC:
      ret += sprintf (& outf [ret], " :\n");
      break;
    case GRS1_SUBTREE:
      if (grs1indent)
	grs1treestr = realloc (grs1treestr, grs1indent + ret1 + 1);
      else
	{
	  grs1treestr = malloc (ret1 + 1);
	  grs1treestr [0] = '\0';
	}
      grs1indent += ret1;
      grs1treestr = strcat (grs1treestr, tmp);

      ret += create_grs1_record ((grs1_genericrecord *) d -> u . subtree,
				 & outf [ret]);
      grs1indent -= ret1;
      if (grs1indent)
	{
	  char * ptr = malloc (grs1indent + 1);
	  memcpy (ptr, grs1treestr, grs1indent);
	  free (grs1treestr);
	  ptr [grs1indent] = '\0';
	  grs1treestr = ptr;
	}
      else
	{
	  free (grs1treestr);
	  grs1treestr = NULL;
	}
      break;
    }

  return (ret);
}


/*
 * Print a GRS.1 Record
 */
static int create_grs1_record (grs1_genericrecord * g, char * outf)
{
  int i;
  int ret = 0;
  int tagval, tagtype, tagocc;
  char * tagstr;

  if (! g || ! g -> tree)
    return (ret);

  for (i = 0; i < g -> size; i ++)
    {
      if (! g -> tree [i])
	continue;

      /* type */
      if (g -> tree [i] -> type)
	tagtype = * g -> tree [i] -> type;
      else
	tagtype = 0;

      /* value */
      if (g -> tree [i] -> value)
	{
	  if (g -> tree [i] -> value -> choice == Z_STRING_STRINGORNUMERIC)
	    {
	      tagstr = g -> tree [i] -> value -> u . string;
	      tagval = 0;
	    }
	  else if (g -> tree [i] -> value -> choice == Z_NUMBER_STRINGORNUMERIC)
	    {
	      tagstr = NULL;
	      tagval = g -> tree [i] -> value -> u . numeric;
	    }
	}

      /* occurrence */
      if (g -> tree [i] -> occurrence)
	tagocc = * g -> tree [i] -> occurrence;
      else
	tagocc = -1;


      ret += create_grs1_data_record (g -> tree [i] -> content,
				      & outf [ret],
				      tagtype, tagval, tagstr, tagocc);
    }

  return (ret);
}



static int create_marc_record (char * buf, char * outf)
{
  int entry_p = 0;
  int record_length = 0;
  int indicator_length = 0;
  int identifier_length = 0;
  int base_address = 0;
  int length_data_entry = 0;
  int length_starting = 0;
  int length_implementation = 0;
  int buf_len = 0;

  int data_length;
  int data_offset;
  int end_offset;
  int i, j;
  char tag [4];


  record_length = atoi_n (buf, 5);
  if (record_length < 25)
    return (-1);

  indicator_length = atoi_n (buf + 10, 1);
  identifier_length = atoi_n (buf + 11, 1);
  base_address = atoi_n (buf + 12, 4);

  length_data_entry = atoi_n (buf + 20, 1);
  length_starting = atoi_n (buf + 21, 1);
  length_implementation = atoi_n (buf + 22, 1);

  for (entry_p = 24; buf [entry_p] != ISO2709_FS;)
    entry_p += 3 + length_data_entry + length_starting;

  base_address = entry_p + 1;
  for (entry_p = 24; buf [entry_p] != ISO2709_FS;)
    {
      memcpy (& outf [buf_len], buf + entry_p, 3);
      buf_len += 3;
      memcpy (tag, buf + entry_p, 3);
      entry_p += 3;
      outf [buf_len ++] = ' ';
      data_length = atoi_n (buf + entry_p, length_data_entry);
      entry_p += length_data_entry;

      data_offset = atoi_n (buf + entry_p, length_starting);
      entry_p += length_starting;

      i = data_offset + base_address;
      end_offset = i + data_length - 1;

      if (memcmp (tag, "00", 2) && indicator_length)
	{
	  memcpy (& outf [buf_len], & buf [i], indicator_length);
	  i += indicator_length;
	  buf_len += indicator_length;
	}

      while (buf[i] != ISO2709_RS && buf[i] != ISO2709_FS && i < end_offset)
	{
	  if (memcmp (tag, "00", 2) && identifier_length)
	    {
	      i ++;
	      memcpy (& outf [buf_len], " $", 2);
              buf_len += 2;
	      memcpy (& outf [buf_len], & buf [i], identifier_length - 1);
	      i += identifier_length - 1;
	      buf_len += identifier_length - 1;
	      outf [buf_len ++]= ' ';
	      while (buf [i] != ISO2709_RS && buf [i] != ISO2709_IDFS &&
		     buf [i] != ISO2709_FS && i < end_offset)
		outf [buf_len ++] = buf [i ++];
	    }
	  else
	    outf [buf_len ++] = buf [i ++];
	}
      outf [buf_len ++] = '\n';

    }

  return (buf_len);
}



static int marclen (char * buf)
{
  int entry_p = 0;
  int record_length = 0;
  int indicator_length = 0;
  int identifier_length = 0;
  int base_address = 0;
  int length_data_entry = 0;
  int length_starting = 0;
  int length_implementation = 0;
  int buf_len = 0;

  int data_length;
  int data_offset;
  int end_offset;
  int i, j;
  char tag [4];


  record_length = atoi_n (buf, 5);
  if (record_length < 25)
    return (-1);

  indicator_length = atoi_n (buf + 10, 1);
  identifier_length = atoi_n (buf + 11, 1);
  base_address = atoi_n (buf + 12, 4);

  length_data_entry = atoi_n (buf + 20, 1);
  length_starting = atoi_n (buf + 21, 1);
  length_implementation = atoi_n (buf + 22, 1);

  for (entry_p = 24; buf [entry_p] != ISO2709_FS;)
    entry_p += 3 + length_data_entry + length_starting;

  base_address = entry_p + 1;
  for (entry_p = 24; buf [entry_p] != ISO2709_FS;)
    {
      memcpy (tag, buf + entry_p, 3);
      entry_p += 3;
      buf_len += 4;
      data_length = atoi_n (buf + entry_p, length_data_entry);
      entry_p += length_data_entry;

      data_offset = atoi_n (buf + entry_p, length_starting);
      entry_p += length_starting;

      i = data_offset + base_address;
      end_offset = i + data_length - 1;

      if (memcmp (tag, "00", 2) && indicator_length)
	{
	  i += indicator_length;
	  buf_len += indicator_length;
	}

      while (buf[i] != ISO2709_RS && buf[i] != ISO2709_FS && i < end_offset)
	{
	  if (memcmp (tag, "00", 2) && identifier_length)
	    {
              buf_len += 1;
	      i += identifier_length;
	      buf_len += identifier_length;
	      buf_len++;
	      for (; buf [i] != ISO2709_RS && buf [i] != ISO2709_IDFS &&
		     buf [i] != ISO2709_FS && i < end_offset; buf_len++, i ++);
	    }
	  else
	    {
	      buf_len ++;
	      i ++;
	    }
	}
      buf_len ++;

    }
  return (buf_len);
}


/*
 * External to ascii
 */
char * external2a (zexternal * e)
{
  char * ciccia = NULL;
  char * ptr;
  int len;
  int i;
  int len1;

  if (! e)
    return (ciccia);

  if (e -> direct_reference)
    len = strlen (e -> direct_reference) + 1;
  else
    len = 1;

  switch (e -> choice)
    {
    case Z_SINGLE_ASN1_TYPE_ENCODING:
      if (e -> direct_reference && ! strcmp (e -> direct_reference, Z_SUTRS))
	{
	  /*
	   * Decode the lenght (drop the Universal TAG 0x1b in first byte)
	   */
	  len1 = sutrslen (e -> encoding . single_asn1_type -> data + 1);
	  if (len1 == -1)
	    return (ciccia);
	  else
	    ptr = e -> encoding . single_asn1_type -> data +
	      e -> encoding . single_asn1_type -> size - len1;
	  /*
	   * buy memory && copy the `ciccia'
	   */
	  ciccia = malloc (len + 1 + len1);
	  memcpy (ciccia, e -> direct_reference,
		  strlen (e -> direct_reference));

	  len = strlen (e -> direct_reference);
	  ciccia [len ++] = '&';
	  memcpy (& ciccia [len], ptr, len1);
	  len += len1;
	}
      else if (e -> direct_reference &&
	       ! strcmp (e -> direct_reference, Z_GRS_1))
	{
	  grs1_genericrecord * grs1 = malloc (sizeof (grs1_genericrecord));
	  int error;

	  /*
	   * Decode the GRS.1 Record
	   */
	  if (decode_grs1 (e -> encoding . single_asn1_type  -> data,
			   e -> encoding . single_asn1_type  -> size,
			   grs1, & error) == -1)
	    {
	      len += e -> encoding . single_asn1_type  -> size;
	      ciccia = malloc (len + 1);

	      memcpy (ciccia, e -> direct_reference,
		      strlen(e -> direct_reference));

	      len = strlen (e -> direct_reference);
	      ciccia [len ++] = '&';

	      memcpy (& ciccia [len] , e -> encoding . single_asn1_type  -> data,
		      e -> encoding . single_asn1_type  -> size);
	    }
	  else if ((len1 = grs1len (grs1)) > 0)
	    {

	      ciccia = malloc (len+len1 + 1);
	      memcpy (ciccia, e -> direct_reference,
		      strlen(e -> direct_reference));

	      len = strlen (e -> direct_reference);
	      ciccia [len ++] = '&';

	      len += create_grs1_record (grs1, & ciccia [len]);
	      grs1_free (grs1);
	    }
	}
      else
	{
	  len += e -> encoding . single_asn1_type  -> size;
	  ciccia = malloc (len + 1);

	  if (e -> direct_reference)
	    {
	      memcpy (ciccia, e -> direct_reference,
		      strlen (e -> direct_reference));
	      len = strlen (e -> direct_reference);
	    }

	  ciccia [len ++] = '&';

	  memcpy (& ciccia [len] , e -> encoding . single_asn1_type  -> data,
		  e -> encoding . single_asn1_type  -> size);
	}
      ciccia [len] = '\0';
      break;

    case Z_OCTET_ALIGNED_ENCODING:
      if (e -> direct_reference && (! strcmp (e -> direct_reference, Z_USMARC) ||
				    ! strcmp (e -> direct_reference, Z_UNIMARC)))
        {
	  for (i = 0; i < e -> encoding . octet_aligned -> size; i ++)
	    {
	      len +=
		marclen (e -> encoding . octet_aligned -> table [i] -> data);
	    }
	  ciccia = malloc (len + 1);
	  memcpy (ciccia, e -> direct_reference,
		  strlen (e ->direct_reference));

	  len = strlen (e -> direct_reference);
	  ciccia [len ++] = '&';

	  for (i = 0; i < e -> encoding . octet_aligned -> size; i ++)
	    {
	      len += create_marc_record (e -> encoding . octet_aligned -> table [i] -> data,
					 & ciccia [len]);
	    }
	}
      else
        {
	  for (i = 0; i < e -> encoding . octet_aligned -> size; i ++)
	    len +=  e -> encoding . octet_aligned -> table [i] -> size;
	  ciccia = malloc (len + 1);

	  if (e -> direct_reference)
	    {
	      memcpy (ciccia, e -> direct_reference, strlen (e ->direct_reference));
	      len = strlen (e -> direct_reference);
	    }
	  ciccia [len ++] = '&';

	  for (i = 0; i < e -> encoding . octet_aligned -> size; i ++)
	    {
	      ptr = e -> encoding . octet_aligned -> table [i] -> data;
	      memcpy (& ciccia [len], ptr,
		      e -> encoding . octet_aligned -> table [i] -> size);
	      len +=  e -> encoding . octet_aligned -> table [i] -> size;
	    }
	}
      ciccia [len] = '\0';
      break;

    case Z_ARBITRARY_ENCODING:
      break;

    default:
      break;
    }

  return (ciccia);
}


/*
 *
 */
char * buildrecords (zrecords * r)
{
  char * record = NULL;
  char * one = NULL;
  char * p = NULL;
  char buffer [200];
  int len = 0;
  int i;

  if (! r)
    return (record);

  switch (r -> choice)
    {
    case Z_RESPONSERECORDS_RECORDS:
      if (! r -> u . responserecords || ! r -> u . responserecords -> table)
	return (record);
      for (i = 0; i < r -> u . responserecords -> size; i ++)
	{
	  if (! r -> u . responserecords -> table [i] -> record)
	    break;
	  switch (r -> u . responserecords -> table [i] -> record -> choice)
	    {
	    case Z_RETRIEVAL_RECORD:
	      one =
		external2a (r -> u . responserecords -> table [i] ->
			    record -> u . retrievalrecord);
	      if (! one)
		break;

	      len = strlen (one);
	      /*
	       * enlarge buffer
	       */
	      if (! record)
		{
		  record = malloc (len + 1);
		  p = record;
		  strcpy (p, one);
		}
	      else
		{
		  record = realloc (record, strlen (record) + len + 2);
		  p = record + strlen (record);
		  * p ++ = '&';
		  strcpy (p, one);
		  p += len;
		}
	      break;

	    case Z_SURROGATE_DIAGNOSTIC:
	      one =
		buildsurrogate (r -> u . responserecords -> table [i] ->
				record -> u . surrogatediagnostic,
				r -> u . responserecords -> table[i]->name);
	      if (! one)
		continue;

	      len = strlen (one);

	      /*
	       * enlarge buffer
	       */
	      if (! record)
		{
		  record = malloc (len + 1);
		  p = record;
		  strcpy (p, one);
		}
	      else
		{
		  record = realloc (record, strlen (record) + len + 2);
		  p = record + strlen (record);
		  * p ++ = '&';
		  strcpy (p, one);
		  p += len;
		}
	      break;

	    case Z_STARTING_FRAGMENT:
	    case Z_INTERMEDIATE_FRAGMENT:
	    case Z_FINAL_FRAGMENT:
	    default:
	      break;
	    }
	}
      break;

    case Z_NONSURROGATEDIAGNOSTIC_RECORDS:
      record = builddefaultdiag (r -> u . nonsurrogatediagnostic, NULL);
      break;

    case Z_MULTIPLENONSURDIAGNOSTICS_RECORDS:
      record = builddiagnostics (r -> u . multiplenonsurdiagnostics -> size,
				 r -> u . multiplenonsurdiagnostics -> table);
      break;

    default:
      break;
    }

  return (record);
}


/*
 *
 */
char * buildentries (zlistentries * e)
{
  char * entry = NULL;
  int i;
  char * one = NULL;
  char * p = NULL;
  char buffer [200];
  int len = 0;

  if (! e)
    return (entry);

  if (e -> entries)
    {
      for (i = 0; i < e -> entries -> size; i ++)
	{
	  switch (e -> entries -> table [i] -> choice)
	    {
	    case Z_SURROGATEDIAGNOSTIC:
	      one =
		buildsurrogate (e -> entries -> table [i] -> u .
				surrogatediagnostic, NULL);
	      if (! one)
		continue;

	      len = strlen (one);

	      /*
	       * enlarge buffer
	       */
	      if (! entry)
		{
		  entry = malloc (len + 1);
		  p = entry;
		  strcpy (p, one);
		}
	      else
		{
		  entry = realloc (entry, strlen (entry) + len + 2);
		  p = entry + strlen (entry);
		  * p ++ = '&';
		  strcpy (p, one);
		}
	      break;

	    }
	}
    }
  else if (e -> nonsurrogatediagnostics)
    entry = builddiagnostics (e -> nonsurrogatediagnostics -> size,
			      e -> nonsurrogatediagnostics -> table);

  return (entry);
}


/*
 * Add the element pointed to by 'r' to the table of input result sets
 */
void addtoinputresultset (zinputresultsetnames * rstable,
			  zinternationalstring * r)
{
  rstable -> table = realloc (rstable -> table, (rstable -> size + 1) *
			      sizeof (zinputresultsetnames *));
  if (! rstable -> table)
    return;
  rstable -> table [rstable -> size] = malloc (strlen (r) + 1);
  if (! rstable -> table [rstable -> size])
    return;
  strcpy (rstable -> table [rstable -> size], r);
  rstable -> size ++;
}


/*
 * A simple routine to fill in a sort element
 * (only generic type is currently supported).
 */
zsortelement * newsortelem (char * s, zattributesetid * id)
{
  int type;
  int value;
  char * p;
  zsortelement * e = NULL;

  /*
   * Buy memory for the descriptor of the sort element
   */
  e = malloc (sizeof (zsortelement));
  if (! e)
    return (NULL);

  e -> choice = Z_SORT_GENERIC;

  e -> u . generic = malloc (sizeof (zsortkey));
  if (! e -> u . generic)
    {
      free (e);
      return (NULL);
    }

  e -> u . generic -> choice = Z_SORTATTRIBUTES_SORTKEY;
  e -> u . generic -> u . sortattributes = malloc (sizeof (zsortattributes));
  if (! e -> u . generic -> u . sortattributes)
    {
      free (e -> u . generic);
      free (e);
      return (NULL);
    }
  e -> u . generic -> u . sortattributes -> id = malloc (strlen (id) + 1);
  if (! e -> u . generic -> u . sortattributes -> id)
    {
      free (e -> u . generic -> u . sortattributes);
      free (e -> u . generic);
      free (e);
      return (NULL);
    }
  strcpy (e -> u . generic -> u . sortattributes -> id, id);

  e -> u . generic -> u . sortattributes -> list =
    malloc (sizeof (zattributelist));
  if (! e -> u . generic -> u . sortattributes -> list)
    {
      free (e -> u . generic -> u . sortattributes -> id);
      free (e -> u . generic -> u . sortattributes);
      free (e -> u . generic);
      free (e);
      return (NULL);
    }
  e -> u . generic -> u . sortattributes -> list -> size  = 0;
  e -> u . generic -> u . sortattributes -> list -> table = NULL;

  /*
   * Parse the string to find the list of attributes
   */
  p = s;
  for (;;)
    {
      sscanf (p, "%d=%d", & type, & value);
      addtolistofattr (e -> u . generic -> u . sortattributes -> list,
		       newattr (type, value, id));
      p = strchr (p, ',');
      if (! p)
	break;
      p ++;
    }

  return (e);
}


/*
 * Add the element pointed to by 's' to the table of key specification
 */
void addtokeyspec (zsortsequence * kstable, char * s, zattributesetid * id,
		   int relation, int sensitivity)
{
  kstable -> table = realloc (kstable -> table, (kstable -> size + 1) *
			      sizeof (zsortkeyspec *));
  if (! kstable -> table)
    return;
  kstable -> table [kstable -> size] = malloc (sizeof (zsortkeyspec));
  if (! kstable -> table [kstable -> size])
    return;
  kstable -> table [kstable -> size] -> sortelement = newsortelem (s, id);
  if (! kstable -> table [kstable -> size] -> sortelement)
    {
      free (kstable -> table [kstable -> size]);
      return;
    }

  kstable -> table [kstable -> size] -> sortrelation       = relation;
  kstable -> table [kstable -> size] -> casesensitivity    = sensitivity;
  kstable -> table [kstable -> size] -> missingvalueaction = NULL;

  kstable -> size ++;
}



char * tinfo2a (zentry * e)
{
  static char current [512];
  static char term [512];
  zterminfo * t;
  int len;

  strcpy (current, "None (0)");

  if (e && e -> choice == Z_TERMINFO)
    {
      t = e -> u . terminfo;

      if (t && t -> term &&
	  t -> term -> choice == Z_GENERAL_TERM &&
	  t -> term -> u . general &&
	  t -> term -> u . general -> data &&
	  t -> term -> u . general -> size > 0)
	{
	  len = t -> term -> u . general -> size;

	  if (t -> globaloccurrences)
	    {
	      memcpy (term, t -> term -> u . general -> data, len);
	      term [len] = '\0';
	      sprintf (current, "%s (%d)", term, * t -> globaloccurrences);
	    }
	  else
	    {
	      memcpy (current, t -> term -> u . general -> data, len);
	      current [len] = '\0';
	    }
	}
    }
  return (current);
}


/*
 * diagnostic for the search response
 */
char * search_error (zrecords * r)
{
  char * text = NULL;
  char * one = NULL;

  if (! r)
    return (NULL);

  if (r -> choice == Z_NONSURROGATEDIAGNOSTIC_RECORDS)
    text = builddefaultdiag (r -> u . nonsurrogatediagnostic, NULL);
  else if (r -> choice == Z_RESPONSERECORDS_RECORDS)
    {
      /* handle only the surrogate records */
      int i, size;
      for (i = 0; i < r -> u. responserecords -> size; i ++)
	{
	  if ( r -> u . responserecords ->table[i] -> record -> choice
               == Z_SURROGATE_DIAGNOSTIC)
	    {
	      one = buildsurrogate (r -> u . responserecords -> table [i] -> record ->
				    u . surrogatediagnostic, NULL);
	      if (! one)
		continue;

	      size = strlen (one);

	      /*
	       * enlarge buffer
	       */
	      if (! text)
		{
		  text = malloc (size + 1);
		  text [0] = '\0';
		  strcpy (text, one);
		}
	      else
		{
		  text = realloc (text, strlen (text) + size + 2);
		  sprintf (text + strlen (text), "&");
		  strcat (text, one);
		}
	    }
	}
    }

  return (text);
}


/*
 * build surrogate records
 */
char * builddiagnostics (unsigned long num, zdiagrec ** table)
{
  char * record = NULL;
  char * one = NULL;
  int i, size;

  if (! num)
    return (NULL);

  for (size = i = 0; i < num; i++)
    {
      one = buildsurrogate (table [i], NULL);
      if (! one)
	continue;

      size = strlen (one);

      /*
       * enlarge buffer
       */
      if (! record)
	{
	  record = malloc (size + 1);
	  record [0] = '\0';
	  strcpy (record, one);
	}
      else
	{
	  record = realloc (record, strlen (record) + size + 2);
	  sprintf (record + strlen (record), "&");
	  strcat (record, one);
	}
    }
  return (record);
}


/*
 * build surrogate record
 */
static char * builddefaultdiag (zdefaultdiagformat * d, char * name)
{
  char * record = NULL;
  char buffer [200];

  if (name)
    record = malloc (25 + strlen (name));
  else
    record = malloc (15);

  if (name)
    sprintf (record, "In DB [%s] Condition: ", name);
  else
    strcpy (record, "Condition: ");

  if (! strcmp (d -> diagnosticsetid, Z_DIAGSET_BIB_1))
    sprintf (buffer, "[%d] that means: %s\n", d -> condition,
	     z_diagbib12a (d -> condition));
  else
    sprintf (buffer, "[%d]\n",d -> condition);

  record = realloc (record, strlen(record) + strlen(buffer) + 1);
  strcat (record, buffer);

  switch (d -> addinfo. choice)
    {
    case Z_ADDINFO_V2ADDINFO_DEFAULTDIAGFORMAT:
      if (d -> addinfo . u . v2addinfo)
	{
	  sprintf (buffer, "Additional info: [%s]\n",
		   d -> addinfo. u. v2addinfo);
	  record = realloc (record, strlen (record) + strlen (buffer) + 1);
	  strcat (record, buffer);
	}
      break;
    case Z_ADDINFO_V3ADDINFO_DEFAULTDIAGFORMAT:
      if (d -> addinfo . u . v3addinfo)
	{
	  sprintf (buffer, "Additional info: [%s]\n",
		   d -> addinfo . u . v3addinfo);
	  record = realloc (record, strlen (record) + strlen (buffer) + 1);
	  strcat (record, buffer);
	}
      break;
    }

  return (record);
}


/*
 * build surrogate record
 */
static char * buildsurrogate (zdiagrec * d, char * name)
{
  if (d -> choice != Z_DEFAULTFORMAT_DIAGREC)
    return (NULL);
  else
    return (builddefaultdiag (d -> u . defaultformat, name));
}


/*
 * Allocate an Authentication identifier of the given type
 */
zidauthentication * zauthidalloc (int choice)
{
  /*
   * Allocate the descriptor first ....
   */
  zidauthentication * id = malloc (sizeof (zidauthentication));

  if (! id)
    return (id);

  id -> choice = choice;
  switch (choice)
    {
    case Z_AUTHTYPE_OPEN:
      id -> u . open = NULL;
      break;

    case Z_AUTHTYPE_IDPASSWORD:
      id -> u . idpass = malloc (sizeof (zidpassword));
      if (! id -> u . idpass)
	{
	  free (id);
	  id = NULL;
	}
      id -> u . idpass -> groupid = NULL;
      id -> u . idpass -> userid = NULL;
      id -> u . idpass -> password = NULL;
      break;

    case Z_AUTHTYPE_ANONYMOUS:
      id -> u . anonymous = NULL;
      break;

    case Z_AUTHTYPE_OTHER:
      id -> u . other = NULL;
      break;
    }

  return (id);
}


/*
 * Free an Authentication identifier
 */
void zauthidfree (zidauthentication * id)
{
  if (! id)
    return;

  switch (id -> choice)
    {
    case Z_AUTHTYPE_OPEN:
      if (id -> u . open)
	free (id -> u . open);
      break;

    case Z_AUTHTYPE_IDPASSWORD:
      if (id -> u . idpass)
	{
	  if (id -> u . idpass -> groupid)
	    free (id -> u . idpass -> groupid);
	  if (id -> u . idpass -> userid)
	    free (id -> u . idpass -> userid);
	  if (id -> u . idpass -> password)
	    free (id -> u . idpass -> password);
	  free (id -> u . idpass);
	}
      break;

    case Z_AUTHTYPE_ANONYMOUS:
      if (id -> u . anonymous)
	free (id -> u . anonymous);
      break;

    case Z_AUTHTYPE_OTHER:
      if (id -> u . other)
	free (id -> u . other);
      break;
    }
  free (id);
}

